Package: llrem
Title: LLM Relational Event Models
Version: 0.1.1
Authors@R: 
    person("C. Ben", "Gibson", email = "cbengibson@gmail.com",
           role = c("aut", "cre"))
Description: Fit Cox proportional hazards relational event models (REMs),
    including a separable formulation that partitions events into initiation
    and continuation sub-models. Optionally augments REM simulations with
    large language model (LLM) agents that select targets conditioned on
    event history, supporting multiple providers ('OpenAI', 'Anthropic',
    'xAI'/'Grok', 'Google Gemini', 'Ollama', 'AWS Bedrock') through a common
    interface. See Butts (2008) <doi:10.1111/j.1467-9531.2008.00203.x>
    for description of relational event modeling.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.1.0)
Imports: ggplot2, httr2 (>= 1.1.0), Rcpp, survival
LinkingTo: Rcpp
Suggests: igraph, parallel, relevent, testthat (>= 3.0.0)
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-05-13 14:03:13 UTC; cbeng
Author: C. Ben Gibson [aut, cre]
Maintainer: C. Ben Gibson <cbengibson@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-19 07:10:08 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-19 09:48:31 UTC; unix
Archs: llrem.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(coef,rem_clogit)
S3method(print,llm_provider)
S3method(print,rem_cfg)
S3method(print,rem_clogit)
S3method(print,rem_covariates)
S3method(print,rem_sep)
export(attach_covariates_df)
export(compare_rem_models)
export(compute_hazard_surface)
export(compute_seq_stat_means)
export(cov_dyad_static)
export(cov_dyad_temporal)
export(cov_node_static)
export(cov_node_temporal)
export(fit_rem)
export(fit_rem_sep)
export(llm_call)
export(llm_provider_anthropic)
export(llm_provider_bedrock)
export(llm_provider_gemini)
export(llm_provider_grok)
export(llm_provider_mock)
export(llm_provider_ollama)
export(llm_provider_openai)
export(llm_provider_openai_compat)
export(make_behavior_queries)
export(make_query_predict)
export(make_query_roleplay_random)
export(make_query_roleplay_sender)
export(make_stat_means)
export(make_suffstats)
export(mock_strategy_highest_indegree)
export(mock_strategy_max_id)
export(mock_strategy_min_id)
export(prepare_edgelist)
export(rem_cfg)
export(rem_covariates)
export(run_llm_rem)
export(run_multiagent_rem)
importFrom(Rcpp,sourceCpp)
importFrom(stats,aggregate)
importFrom(stats,coef)
importFrom(stats,median)
importFrom(stats,model.matrix)
importFrom(stats,setNames)
importFrom(stats,terms)
useDynLib(llrem, .registration = TRUE)

# Exogenous covariate support for REM risk sets --------------------------------
#
# Four covariate types, each an S3 list with class c("<type>", "rem_cov"):
#   cov_node_static    – one value per node
#   cov_node_temporal  – one value per (node, time-window)
#   cov_dyad_static    – one value per directed dyad
#   cov_dyad_temporal  – one value per (dyad, time-window)
#
# All four are collected into a rem_covariates() container that validates
# names and passes as a single argument through the pipeline.

.reserved_cols <- c(
  "sender", "receiver", "eid", "occurred", "time", "time1", "time2", "eveid",
  "senOutDg", "dyad_occurrence", "Repetition", "Recency", "RecRecencyLog", "RecRecencyRank",
  "PSABAB", "PSABBA", "PSABAY", "PSABXA", "PSABXB", "PSABBY", "PSABXY",
  "IOviolation", "recInDg", "PreImitation", "Imitation",
  "OSP", "ISP", "ITP", "OTP", "NOTPSnd", "NITPSnd", "NOSPSnd", "NISPSnd",
  "triadic_closure", "NTriadic", "SndInitiation", "RRecInitiation", "DyadRecencyLog",
  "w_ji", "WeightedRep", "WeightedRec",
  "OutDegSnd", "OutDegRec", "InDegSnd", "InDegRec", "TotDegSnd", "TotDegRec",
  "FrRecSnd", "senInDg", "recOutDg", "n_norm",
  "PSAB_BA", "PSAB_BY", "PSAB_AY", "PSAB_XA", "PSAB_XB", "PSAB_XY"
)

# Constructors -----------------------------------------------------------------

#' Create a static node-level covariate
#'
#' Adds one or two columns to each risk-set row based on the sender's and/or
#' receiver's value in a node attribute table.
#'
#' @param name Character scalar. Column name(s) added to the risk set.
#'   When `role = "both"`, two columns are added: `<name>_snd` and
#'   `<name>_rec`. Otherwise a single column named `<name>` is added.
#' @param data Data frame with columns `node` (integer-coercible) and
#'   `value` (numeric). One row per node.
#' @param role One of `"both"` (default), `"sender"`, or `"receiver"`.
#' @param warn_missing Logical. Warn when a node in the risk set has no
#'   matching row in `data` (default `TRUE`).
#'
#' @return An object of class `c("cov_node_static", "rem_cov")`.
#' @examples
#' gdp <- data.frame(node = 1:4, value = c(1.2, 0.8, 2.1, 0.5))
#' cov <- cov_node_static("GDP", gdp, role = "sender")
#' @export
cov_node_static <- function(name, data, role = c("both", "sender", "receiver"),
                             warn_missing = TRUE) {
  role <- match.arg(role)
  stopifnot(is.character(name), length(name) == 1L)
  stopifnot(is.data.frame(data))
  if (!all(c("node", "value") %in% names(data)))
    stop("'data' must have columns 'node' and 'value'.", call. = FALSE)
  data$node <- as.integer(data$node)
  structure(
    list(name = name, data = data, role = role, warn_missing = warn_missing),
    class = c("cov_node_static", "rem_cov")
  )
}

#' Create a temporal node-level covariate
#'
#' Like [cov_node_static()] but values change at known time breakpoints.
#' The active value at event time `t` is the one with the largest
#' `time_start <= t` (step function).
#'
#' @param name Character scalar.
#' @param data Data frame with columns `node` (integer-coercible),
#'   `time_start` (numeric), and `value` (numeric). Rows need not be
#'   sorted; sorting is done internally.
#' @param role One of `"both"` (default), `"sender"`, or `"receiver"`.
#' @param warn_missing Logical.
#'
#' @return An object of class `c("cov_node_temporal", "rem_cov")`.
#' @examples
#' sanctions <- data.frame(
#'   node = c(1L, 1L, 2L), time_start = c(0, 10, 5), value = c(0, 1, 1)
#' )
#' cov <- cov_node_temporal("Sanctioned", sanctions, role = "receiver")
#' @export
cov_node_temporal <- function(name, data,
                               role = c("both", "sender", "receiver"),
                               warn_missing = TRUE) {
  role <- match.arg(role)
  stopifnot(is.character(name), length(name) == 1L)
  stopifnot(is.data.frame(data))
  if (!all(c("node", "time_start", "value") %in% names(data)))
    stop("'data' must have columns 'node', 'time_start', and 'value'.",
         call. = FALSE)
  data$node <- as.integer(data$node)
  # Pre-split by node for O(log W) lookup at merge time instead of O(m)
  data       <- data[order(data$node, data$time_start), ]
  by_node    <- split(data[, c("time_start", "value")], data$node)
  structure(
    list(name = name, data = data, by_node = by_node,
         role = role, warn_missing = warn_missing),
    class = c("cov_node_temporal", "rem_cov")
  )
}

#' Create a static dyad-level covariate
#'
#' Adds one column to each risk-set row based on the (sender, receiver) pair.
#'
#' @param name Character scalar. Column name added to the risk set.
#' @param data Data frame with columns `sender` (integer-coercible),
#'   `receiver` (integer-coercible), and `value` (numeric). Need not be a
#'   complete cross-product; unmatched dyads receive 0 (with optional warning).
#' @param warn_missing Logical.
#'
#' @return An object of class `c("cov_dyad_static", "rem_cov")`.
#' @examples
#' alliance <- data.frame(
#'   sender = c(1L, 2L), receiver = c(2L, 1L), value = c(1, 1)
#' )
#' cov <- cov_dyad_static("Alliance", alliance)
#' @export
cov_dyad_static <- function(name, data, warn_missing = TRUE) {
  stopifnot(is.character(name), length(name) == 1L)
  stopifnot(is.data.frame(data))
  if (!all(c("sender", "receiver", "value") %in% names(data)))
    stop("'data' must have columns 'sender', 'receiver', and 'value'.",
         call. = FALSE)
  data$sender   <- as.integer(data$sender)
  data$receiver <- as.integer(data$receiver)
  # Pre-compute eid key once at construction
  data$.eid <- paste0(data$sender, "_", data$receiver)
  structure(
    list(name = name, data = data, warn_missing = warn_missing),
    class = c("cov_dyad_static", "rem_cov")
  )
}

#' Create a temporal dyad-level covariate
#'
#' Like [cov_dyad_static()] but values change at known time breakpoints.
#'
#' @param name Character scalar.
#' @param data Data frame with columns `sender`, `receiver`, `time_start`,
#'   and `value`. Rows need not be sorted.
#' @param warn_missing Logical.
#'
#' @return An object of class `c("cov_dyad_temporal", "rem_cov")`.
#' @examples
#' treaty <- data.frame(
#'   sender = c(1L, 1L), receiver = c(2L, 2L),
#'   time_start = c(0, 15), value = c(0, 1)
#' )
#' cov <- cov_dyad_temporal("Treaty", treaty)
#' @export
cov_dyad_temporal <- function(name, data, warn_missing = TRUE) {
  stopifnot(is.character(name), length(name) == 1L)
  stopifnot(is.data.frame(data))
  if (!all(c("sender", "receiver", "time_start", "value") %in% names(data)))
    stop("'data' must have columns 'sender', 'receiver', 'time_start', and 'value'.",
         call. = FALSE)
  data$sender   <- as.integer(data$sender)
  data$receiver <- as.integer(data$receiver)
  data$.eid     <- paste0(data$sender, "_", data$receiver)
  data          <- data[order(data$.eid, data$time_start), ]
  by_dyad       <- split(data[, c("time_start", "value")], data$.eid)
  structure(
    list(name = name, data = data, by_dyad = by_dyad, warn_missing = warn_missing),
    class = c("cov_dyad_temporal", "rem_cov")
  )
}

#' Collect covariate objects for use in fit_rem
#'
#' Validates that all covariate names are unique, do not collide with
#' endogenous REM column names, and are proper `rem_cov` objects.
#'
#' @param ... Any number of objects created by [cov_node_static()],
#'   [cov_node_temporal()], [cov_dyad_static()], or [cov_dyad_temporal()].
#'
#' @return An object of class `"rem_covariates"`.
#' @examples
#' gdp  <- cov_node_static("GDP", data.frame(node = 1:4, value = c(1,2,1,3)))
#' ally <- cov_dyad_static("Alliance",
#'           data.frame(sender = 1L, receiver = 2L, value = 1))
#' covs <- rem_covariates(gdp, ally)
#' @export
rem_covariates <- function(...) {
  covs <- list(...)
  if (length(covs) == 0L)
    return(structure(list(covariates = list()), class = "rem_covariates"))

  for (i in seq_along(covs)) {
    if (!inherits(covs[[i]], "rem_cov"))
      stop("All arguments to rem_covariates() must be rem_cov objects ",
           "(created by cov_node_static, cov_node_temporal, ",
           "cov_dyad_static, or cov_dyad_temporal).", call. = FALSE)
  }

  all_names <- vapply(covs, `[[`, character(1L), "name")

  # node covariates with role="both" produce two columns: name_snd, name_rec
  col_names <- unlist(lapply(covs, function(cv) {
    if (inherits(cv, c("cov_node_static", "cov_node_temporal")) &&
        cv$role == "both") {
      c(paste0(cv$name, "_snd"), paste0(cv$name, "_rec"))
    } else {
      cv$name
    }
  }))

  dupes <- col_names[duplicated(col_names)]
  if (length(dupes))
    stop("Duplicate covariate column name(s): ",
         paste(unique(dupes), collapse = ", "), call. = FALSE)

  all_to_check  <- unique(c(all_names, col_names))
  reserved_hits <- all_to_check[all_to_check %in% .reserved_cols]
  if (length(reserved_hits))
    stop("Covariate name(s) collide with reserved REM columns: ",
         paste(reserved_hits, collapse = ", "), call. = FALSE)

  structure(list(covariates = covs), class = "rem_covariates")
}

#' @export
print.rem_covariates <- function(x, ...) {
  covs <- x$covariates
  if (length(covs) == 0L) {
    cat("rem_covariates: (empty)\n")
    return(invisible(x))
  }
  cat("rem_covariates with", length(covs), "covariate(s):\n")
  for (cv in covs) {
    type <- sub("cov_", "", class(cv)[1L])
    extra <- if (inherits(cv, c("cov_node_static", "cov_node_temporal")))
      paste0(", role=", cv$role) else ""
    cat(sprintf("  %-20s  [%s%s]\n", cv$name, type, extra))
  }
  invisible(x)
}

# Internal merge worker --------------------------------------------------------

# Step-function lookup: index of last time_start <= t, or 0 if none.
.find_window <- function(t, time_starts) {
  findInterval(t, time_starts, rightmost.closed = FALSE, left.open = FALSE)
}

# Look up node values for a vector of node IDs given one covariate and event time.
.node_vals <- function(nodes, cv, t) {
  if (inherits(cv, "cov_node_static")) {
    idx <- match(nodes, cv$data$node)
    cv$data$value[idx]
  } else {
    # temporal: per-node step-function lookup
    vals <- vapply(nodes, function(nd) {
      d <- cv$by_node[[as.character(nd)]]
      if (is.null(d)) return(NA_real_)
      w <- .find_window(t, d$time_start)
      if (w == 0L) NA_real_ else d$value[w]
    }, numeric(1L))
    vals
  }
}

# Append all covariates in a rem_covariates object to a riskset data frame.
# event_time: scalar integer/numeric clock time of the current event.
attach_covariates <- function(riskset, covariates, event_time) {
  if (is.null(covariates) || length(covariates$covariates) == 0L)
    return(riskset)

  for (cv in covariates$covariates) {

    if (inherits(cv, c("cov_node_static", "cov_node_temporal"))) {
      if (cv$role %in% c("sender", "both")) {
        vals <- .node_vals(riskset$sender, cv, event_time)
        n_miss <- sum(is.na(vals))
        if (n_miss > 0L && cv$warn_missing)
          warning(sprintf("cov '%s' (_snd): %d risk-set row(s) have no node match",
                          cv$name, n_miss), call. = FALSE)
        col <- if (cv$role == "both") paste0(cv$name, "_snd") else cv$name
        riskset[[col]] <- vals
      }
      if (cv$role %in% c("receiver", "both")) {
        vals <- .node_vals(riskset$receiver, cv, event_time)
        n_miss <- sum(is.na(vals))
        if (n_miss > 0L && cv$warn_missing)
          warning(sprintf("cov '%s' (_rec): %d risk-set row(s) have no node match",
                          cv$name, n_miss), call. = FALSE)
        col <- if (cv$role == "both") paste0(cv$name, "_rec") else cv$name
        riskset[[col]] <- vals
      }

    } else if (inherits(cv, "cov_dyad_static")) {
      idx  <- match(riskset$eid, cv$data$.eid)
      vals <- cv$data$value[idx]
      n_miss <- sum(is.na(vals))
      if (n_miss > 0L && cv$warn_missing)
        warning(sprintf("cov '%s': %d risk-set row(s) have no dyad match",
                        cv$name, n_miss), call. = FALSE)
      riskset[[cv$name]] <- vals

    } else if (inherits(cv, "cov_dyad_temporal")) {
      vals <- vapply(riskset$eid, function(eid) {
        d <- cv$by_dyad[[eid]]
        if (is.null(d)) return(NA_real_)
        w <- .find_window(event_time, d$time_start)
        if (w == 0L) NA_real_ else d$value[w]
      }, numeric(1L))
      n_miss <- sum(is.na(vals))
      if (n_miss > 0L && cv$warn_missing)
        warning(sprintf("cov '%s': %d risk-set row(s) have no dyad/time match",
                        cv$name, n_miss), call. = FALSE)
      riskset[[cv$name]] <- vals
    }
  }
  riskset
}

#' Attach exogenous covariates to a pre-built risk-set data frame
#'
#' Convenience wrapper for applying a [rem_covariates()] object to the output
#' of `remstats_generate()` after the fact.  Useful for testing or for
#' workflows that build the risk set once and experiment with different
#' covariate specifications.
#'
#' @param pe          Data frame returned by `remstats_generate()` (or
#'   equivalently, the `pe` object inside [fit_rem()]), with a `time` column.
#' @param covariates  A [rem_covariates()] object.
#'
#' @return `pe` with additional covariate columns appended.
#' @examples
#' \dontrun{
#' data(edgelist)
#' el   <- prepare_edgelist(edgelist)
#' gdp  <- cov_node_static("GDP", data.frame(
#'   node = sort(unique(c(el$sender, el$receiver))),
#'   value = runif(length(unique(c(el$sender, el$receiver))))
#' ))
#' covs <- rem_covariates(gdp)
#' fit  <- fit_rem(el[1:100, ],
#'   survival::Surv(time1, occurred) ~ GDP + Repetition +
#'     survival::strata(time2),
#'   n_nonevents = 30L, method = "clogit", covariates = covs)
#' }
#' @export
attach_covariates_df <- function(pe, covariates) {
  if (is.null(covariates) || length(covariates$covariates) == 0L)
    return(pe)
  chunks <- split(pe, pe$time)
  chunks <- lapply(chunks, function(chunk) {
    t <- chunk$time[1L]
    attach_covariates(chunk, covariates, t)
  })
  do.call(rbind, chunks)
}

# Tests for rem_covariates and covariate merging

# ---- constructors and validation --------------------------------------------

test_that("cov_node_static constructs and validates", {
  d <- data.frame(node = 1:3, value = c(1, 2, 3))
  cv <- cov_node_static("x", d)
  expect_s3_class(cv, "cov_node_static")
  expect_s3_class(cv, "rem_cov")
  expect_equal(cv$name, "x")
  expect_equal(cv$role, "both")
})

test_that("cov_node_static errors on bad data", {
  expect_error(cov_node_static("x", data.frame(a = 1, b = 2)), "node")
})

test_that("cov_node_temporal constructs and pre-sorts by node + time_start", {
  d <- data.frame(node = c(1L, 1L, 2L), time_start = c(5, 1, 1),
                  value = c(10, 5, 20))
  cv <- cov_node_temporal("y", d)
  expect_s3_class(cv, "cov_node_temporal")
  expect_equal(cv$by_node[["1"]]$time_start, c(1, 5))
})

test_that("cov_dyad_static constructs and pre-computes eid key", {
  d <- data.frame(sender = c(1L, 2L), receiver = c(2L, 1L), value = c(0.5, 0.3))
  cv <- cov_dyad_static("dist", d)
  expect_s3_class(cv, "cov_dyad_static")
  expect_true(".eid" %in% names(cv$data))
  expect_equal(cv$data$.eid, c("1_2", "2_1"))
})

test_that("cov_dyad_temporal constructs and pre-splits by dyad", {
  d <- data.frame(sender = c(1L, 1L), receiver = c(2L, 2L),
                  time_start = c(1, 10), value = c(0.1, 0.9))
  cv <- cov_dyad_temporal("trade", d)
  expect_s3_class(cv, "cov_dyad_temporal")
  expect_true("1_2" %in% names(cv$by_dyad))
})

test_that("rem_covariates validates name uniqueness", {
  d <- data.frame(node = 1:2, value = 1:2)
  expect_error(
    rem_covariates(cov_node_static("x", d), cov_node_static("x", d)),
    "Duplicate"
  )
})

test_that("rem_covariates blocks reserved column names", {
  d <- data.frame(node = 1:2, value = 1:2)
  expect_error(rem_covariates(cov_node_static("Repetition", d)), "reserved")
})

test_that("rem_covariates with role=both blocks <name>_snd collision", {
  d <- data.frame(node = 1:2, value = 1:2)
  # _snd suffix would collide if two covariates expand to same name
  cv1 <- cov_node_static("gdp", d, role = "sender")   # column: gdp
  cv2 <- cov_node_static("gdp", d, role = "receiver")  # column: gdp -> duplicate
  expect_error(rem_covariates(cv1, cv2), "Duplicate")
})

test_that("empty rem_covariates is valid", {
  rc <- rem_covariates()
  expect_s3_class(rc, "rem_covariates")
  expect_length(rc$covariates, 0L)
})

# ---- attach_covariates merge correctness ------------------------------------

make_rs <- function() {
  data.frame(
    sender   = c(1L, 1L, 2L, 2L),
    receiver = c(2L, 3L, 1L, 3L),
    eid      = c("1_2", "1_3", "2_1", "2_3"),
    stringsAsFactors = FALSE
  )
}

test_that("cov_node_static role=both adds _snd and _rec columns", {
  rs <- make_rs()
  d  <- data.frame(node = 1:3, value = c(10, 20, 30))
  rc <- rem_covariates(cov_node_static("pop", d, warn_missing = FALSE))
  out <- llrem:::attach_covariates(rs, rc, event_time = 1)
  expect_true("pop_snd" %in% names(out))
  expect_true("pop_rec" %in% names(out))
  # row 1: sender=1 -> pop_snd=10, receiver=2 -> pop_rec=20
  expect_equal(out$pop_snd, c(10, 10, 20, 20))
  expect_equal(out$pop_rec, c(20, 30, 10, 30))
})

test_that("cov_node_static role=sender adds one column named <name>", {
  rs  <- make_rs()
  d   <- data.frame(node = 1:3, value = c(1, 2, 3))
  rc  <- rem_covariates(cov_node_static("x", d, role = "sender",
                                         warn_missing = FALSE))
  out <- llrem:::attach_covariates(rs, rc, event_time = 1)
  expect_true("x" %in% names(out))
  expect_false("x_snd" %in% names(out))
  expect_equal(out$x, c(1, 1, 2, 2))
})

test_that("missing node match keeps NA and warns", {
  rs <- make_rs()
  d  <- data.frame(node = 1L, value = 5)   # only node 1 present
  rc <- rem_covariates(cov_node_static("x", d, role = "sender"))
  expect_warning(
    out <- llrem:::attach_covariates(rs, rc, event_time = 1),
    "no node match"
  )
  # node 2 has no match -> NA preserved
  expect_equal(out$x, c(5, 5, NA_real_, NA_real_))
})

test_that("cov_node_temporal step-function respects time windows", {
  rs <- make_rs()
  d  <- data.frame(
    node       = c(1L, 1L, 2L, 2L, 3L),
    time_start = c(1,  10,  1,  10,  1),
    value      = c(1,   2,  3,   4,  5)
  )
  rc <- rem_covariates(cov_node_temporal("x", d, role = "sender",
                                          warn_missing = FALSE))
  # event at t=5: nodes 1 and 2 in window 1 (time_start=1)
  out5 <- llrem:::attach_covariates(rs, rc, event_time = 5)
  expect_equal(out5$x, c(1, 1, 3, 3))

  # event at t=15: nodes 1 and 2 in window 2 (time_start=10)
  out15 <- llrem:::attach_covariates(rs, rc, event_time = 15)
  expect_equal(out15$x, c(2, 2, 4, 4))
})

test_that("cov_dyad_static fills matched values and keeps NA for unmatched", {
  rs <- make_rs()
  d  <- data.frame(sender = c(1L, 2L), receiver = c(2L, 3L), value = c(100, 200))
  rc <- rem_covariates(cov_dyad_static("dist", d, warn_missing = FALSE))
  out <- llrem:::attach_covariates(rs, rc, event_time = 1)
  # 1_2 -> 100, 1_3 -> NA, 2_1 -> NA, 2_3 -> 200
  expect_equal(out$dist, c(100, NA_real_, NA_real_, 200))
})

test_that("cov_dyad_temporal step-function lookup is correct", {
  rs <- make_rs()
  d  <- data.frame(
    sender     = c(1L, 1L),
    receiver   = c(2L, 2L),
    time_start = c(1,  10),
    value      = c(0.1, 0.9)
  )
  rc <- rem_covariates(cov_dyad_temporal("trade", d, warn_missing = FALSE))
  out7  <- llrem:::attach_covariates(rs, rc, event_time = 7)
  out12 <- llrem:::attach_covariates(rs, rc, event_time = 12)
  expect_equal(out7$trade[out7$eid == "1_2"],  0.1)
  expect_equal(out12$trade[out12$eid == "1_2"], 0.9)
})

# ---- integration: fit_rem with covariates -----------------------------------

test_that("fit_rem clogit with a node covariate runs and coefficient is finite", {
  el  <- make_edgelist(n_nodes = 6L, n_events = 25L, seed = 99L)
  rs  <- llrem:::get_base_riskset(NULL, llrem:::get_nodes(el))

  # simple static covariate: node "size" 1-6
  d   <- data.frame(node = 1:6, value = as.numeric(1:6))
  rc  <- rem_covariates(cov_node_static("size", d, warn_missing = FALSE))

  f   <- survival::Surv(time1, occurred) ~
    Repetition + RecRecencyLog + IOviolation + size_snd + size_rec +
    survival::strata(time2)

  suppressWarnings(
    m <- fit_rem(el, f, n_nonevents = NULL, base_riskset = rs,
                 method = "clogit", covariates = rc)
  )
  expect_true(is.finite(m$coef["size_snd"]))
  expect_true(is.finite(m$coef["size_rec"]))
})

test_that("fit_rem clogit with a dyad covariate runs and coefficient is finite", {
  el    <- make_edgelist(n_nodes = 6L, n_events = 25L, seed = 42L)
  rs    <- llrem:::get_base_riskset(NULL, llrem:::get_nodes(el))
  nodes <- llrem:::get_nodes(el)

  # dyad covariate: random distance between all pairs
  set.seed(7)
  dyad_d <- expand.grid(sender = nodes, receiver = nodes)
  dyad_d <- dyad_d[dyad_d$sender != dyad_d$receiver, ]
  dyad_d$value <- runif(nrow(dyad_d))
  rc  <- rem_covariates(cov_dyad_static("dist", dyad_d, warn_missing = FALSE))

  f   <- survival::Surv(time1, occurred) ~
    Repetition + dist + survival::strata(time2)

  suppressWarnings(
    m <- fit_rem(el, f, n_nonevents = NULL, base_riskset = rs,
                 method = "clogit", covariates = rc)
  )
  expect_true(is.finite(m$coef["dist"]))
})

test_that("endogenous NAs are filled to 0; exogenous NAs are preserved", {
  el  <- make_edgelist(n_nodes = 6L, n_events = 15L, seed = 5L)
  rs  <- llrem:::get_base_riskset(NULL, llrem:::get_nodes(el))
  # covariate only covers nodes 1-3, leaving nodes 4-6 as NA
  d   <- data.frame(node = 1:3, value = c(10, 20, 30))
  rc  <- rem_covariates(cov_node_static("pop", d, role = "sender",
                                         warn_missing = FALSE))
  suppressWarnings(
    pe <- llrem:::remstats_generate(el, n_nonevents = NULL,
                                    base_riskset = rs, covariates = rc)
  )
  # endogenous: Repetition should have no NAs
  expect_false(anyNA(pe$Repetition))
  # exogenous: pop should have NAs for senders 4-6
  expect_true(anyNA(pe$pop))
  expect_true(all(!is.na(pe$pop[pe$sender %in% 1:3])))
  expect_true(all( is.na(pe$pop[pe$sender %in% 4:6])))
})

test_that("NULL covariates produces identical output to no covariates argument", {
  el <- make_edgelist(n_nodes = 5L, n_events = 15L, seed = 1L)
  rs <- llrem:::get_base_riskset(NULL, llrem:::get_nodes(el))
  f  <- survival::Surv(time1, occurred) ~
    Repetition + RecRecencyLog + survival::strata(time2)

  suppressWarnings({
    m1 <- fit_rem(el, f, n_nonevents = NULL, base_riskset = rs, method = "clogit")
    m2 <- fit_rem(el, f, n_nonevents = NULL, base_riskset = rs, method = "clogit",
                  covariates = NULL)
    m3 <- fit_rem(el, f, n_nonevents = NULL, base_riskset = rs, method = "clogit",
                  covariates = rem_covariates())
  })
  expect_equal(m1$coef, m2$coef)
  expect_equal(m1$coef, m3$coef)
})

# Shared synthetic data and config used across test files.

make_edgelist <- function(n_nodes = 6L, n_events = 20L, seed = 1776L) {
  set.seed(seed)
  out <- data.frame(time = integer(n_events), sender = integer(n_events),
                    receiver = integer(n_events))
  for (i in seq_len(n_events)) {
    s <- sample.int(n_nodes, 1L)
    r <- sample(setdiff(seq_len(n_nodes), s), 1L)
    out[i, ] <- list(i, s, r)
  }
  prepare_edgelist(out)
}

test_edgelist  <- make_edgelist()
test_nodes     <- get_nodes(test_edgelist)
test_riskset   <- get_base_riskset(NULL, test_nodes)
test_formula   <- survival::Surv(time = time1, event = occurred) ~
  Imitation + Repetition + SndInitiation + RecRecencyLog + IOviolation + NTriadic +
  survival::strata(time2)

mock_cfg <- rem_cfg(
  provider       = llm_provider_mock(mock_strategy_highest_indegree),
  context        = "A small test network.",
  action_verb    = "contacted",
  actor_noun     = "Node",
  history_window = NULL
)

# Internal constructor ---------------------------------------------------------

new_provider <- function(name, model, call_fn) {
  structure(list(name = name, model = model, call = call_fn),
            class = "llm_provider")
}

# Parse the Retry-After header value (seconds) from a response, or return NULL.
retry_after_secs <- function(resp) {
  h <- httr2::resp_header(resp, "Retry-After")
  if (!is.null(h)) {
    n <- suppressWarnings(as.numeric(h))
    if (!is.na(n) && n > 0) {
      message("[llrem] Retry-After: ", n, "s")
      return(n)
    }
  }
  NULL  # fall through to backoff
}

# Execute an httr2 request, catching only network-level errors.
# Rate-limit and server errors are handled by req_retry inside the request.
execute_req <- function(req, extract_fn, max_retries = 3L) {
  for (attempt in seq_len(max_retries)) {
    resp <- tryCatch(
      httr2::req_perform(req),
      error = function(e) {
        message("[llrem] Request error (attempt ", attempt, "): ",
                conditionMessage(e))
        NULL
      }
    )
    if (!is.null(resp) && httr2::resp_status(resp) == 200L)
      return(extract_fn(resp))
    if (attempt < max_retries) Sys.sleep(min(2^(attempt + 2L), 120L))
  }
  warning("[llrem] LLM request failed after ", max_retries,
          " attempts; returning NA.", call. = FALSE)
  NA_character_
}

# Exported constructors -------------------------------------------------------

#' @export
print.llm_provider <- function(x, ...) {
  cat("<llm_provider>", x$name, "/", x$model, "\n")
  invisible(x)
}

#' Create a generic OpenAI-compatible LLM provider
#'
#' Suitable for any API that mirrors the OpenAI chat completions wire format.
#' Used internally by [llm_provider_openai()], [llm_provider_grok()],
#' [llm_provider_gemini()], and [llm_provider_ollama()].
#'
#' @param api_key API key string.
#' @param model Model identifier.
#' @param base_url Full chat-completions endpoint URL.
#' @param name Display name used in `print()` output and plot labels.
#' @param max_retries Maximum request attempts before returning `NA`.
#'
#' @return An `llm_provider` object.
#' @examples
#' \dontrun{
#' p <- llm_provider_openai_compat(
#'   api_key  = Sys.getenv("OPENAI_API_KEY"),
#'   model    = "gpt-4o-mini",
#'   base_url = "https://api.openai.com/v1/chat/completions"
#' )
#' print(p)
#' }
#' @export
llm_provider_openai_compat <- function(api_key,
                                        model,
                                        base_url,
                                        name        = "OpenAI-compatible",
                                        max_retries = 5L) {
  force(api_key); force(model); force(base_url); force(name); force(max_retries)
  call_fn <- function(messages, max_tokens = 15L, temperature = 0.7) {
    req <- httr2::request(base_url) |>
      httr2::req_auth_bearer_token(api_key) |>
      httr2::req_headers("Content-Type" = "application/json") |>
      httr2::req_body_json(list(
        model       = model,
        messages    = messages,
        max_tokens  = max_tokens,
        temperature = temperature
      )) |>
      httr2::req_retry(
        max_tries    = max_retries,
        is_transient = \(r) httr2::resp_status(r) %in% c(429L, 500L, 503L),
        backoff      = \(i) min(2^i, 60L),
        after        = retry_after_secs
      )
    execute_req(req, \(resp) {
      trimws(httr2::resp_body_json(resp)$choices[[1L]]$message$content)
    }, max_retries = 3L)
  }
  new_provider(name, model, call_fn)
}

#' Create an OpenAI LLM provider
#'
#' @param api_key Defaults to `Sys.getenv("OPENAI_API_KEY")`.
#' @param model Default `"gpt-4o"`.
#' @param max_retries Maximum request attempts.
#'
#' @return An `llm_provider` object.
#' @examples
#' \dontrun{
#' p <- llm_provider_openai()   # reads OPENAI_API_KEY from env
#' print(p)
#' }
#' @export
llm_provider_openai <- function(api_key     = Sys.getenv("OPENAI_API_KEY"),
                                 model       = "gpt-4o",
                                 max_retries = 5L) {
  llm_provider_openai_compat(api_key, model,
                               "https://api.openai.com/v1/chat/completions",
                               name = "OpenAI", max_retries = max_retries)
}

#' Create an xAI Grok LLM provider
#'
#' @param api_key Defaults to `Sys.getenv("XAI_API_KEY")`.
#' @param model Default `"grok-3"`.
#' @param max_retries Maximum request attempts.
#'
#' @return An `llm_provider` object.
#' @examples
#' \dontrun{
#' p <- llm_provider_grok()   # reads XAI_API_KEY from env
#' print(p)
#' }
#' @export
llm_provider_grok <- function(api_key     = Sys.getenv("XAI_API_KEY"),
                               model       = "grok-3",
                               max_retries = 5L) {
  llm_provider_openai_compat(api_key, model,
                               "https://api.x.ai/v1/chat/completions",
                               name = "xAI/Grok", max_retries = max_retries)
}

#' Create a Google Gemini LLM provider
#'
#' Uses Google's OpenAI-compatible endpoint.
#'
#' @param api_key Defaults to `Sys.getenv("GEMINI_API_KEY")`.
#' @param model Default `"gemini-2.0-flash"`.
#' @param max_retries Maximum request attempts.
#'
#' @return An `llm_provider` object.
#' @examples
#' \dontrun{
#' p <- llm_provider_gemini()   # reads GEMINI_API_KEY from env
#' print(p)
#' }
#' @export
llm_provider_gemini <- function(api_key     = Sys.getenv("GEMINI_API_KEY"),
                                 model       = "gemini-2.0-flash",
                                 max_retries = 5L) {
  llm_provider_openai_compat(
    api_key, model,
    "https://generativelanguage.googleapis.com/v1beta/openai/chat/completions",
    name = "Google/Gemini", max_retries = max_retries
  )
}

#' Create a local or remote Ollama LLM provider
#'
#' @param model       Model tag, e.g. `"llama3.2"` or `"mistral"`.
#' @param host        Hostname or IP of the Ollama server.  Defaults to
#'   `"localhost"`.  Override to target a remote machine, e.g.
#'   `host = "192.168.1.65"`.
#' @param port        Port the Ollama server is listening on (default 11434).
#' @param base_url    Full endpoint URL.  When supplied, `host` and `port` are
#'   ignored — useful as a manual escape hatch.
#' @param max_retries Maximum request attempts.
#'
#' @return An `llm_provider` object.
#' @examples
#' \dontrun{
#' p <- llm_provider_ollama(model = "llama3.2")   # requires local Ollama server
#' print(p)
#' }
#' @export
llm_provider_ollama <- function(model       = "llama3.2",
                                 host        = "localhost",
                                 port        = 11434L,
                                 base_url    = NULL,
                                 max_retries = 3L) {
  url <- if (!is.null(base_url)) base_url else
    sprintf("http://%s:%d/v1/chat/completions", host, as.integer(port))
  llm_provider_openai_compat("ollama", model, url,
                               name = "Ollama", max_retries = max_retries)
}

#' Create an Anthropic Claude LLM provider
#'
#' Anthropic's API separates the system prompt as a top-level field and uses a
#' different response shape than the OpenAI format. This constructor handles
#' both translations internally.
#'
#' @param api_key Defaults to `Sys.getenv("ANTHROPIC_API_KEY")`.
#' @param model Default `"claude-3-5-sonnet-20241022"`.
#' @param anthropic_version API version header string.
#' @param max_retries Maximum request attempts.
#'
#' @return An `llm_provider` object.
#' @examples
#' \dontrun{
#' p <- llm_provider_anthropic()   # reads ANTHROPIC_API_KEY from env
#' print(p)
#' }
#' @export
llm_provider_anthropic <- function(api_key           = Sys.getenv("ANTHROPIC_API_KEY"),
                                    model             = "claude-sonnet-4-6",
                                    anthropic_version = "2023-06-01",
                                    max_retries       = 5L) {
  force(api_key); force(model); force(anthropic_version); force(max_retries)
  call_fn <- function(messages, max_tokens = 15L, temperature = 0.7) {
    is_sys   <- vapply(messages, \(m) identical(m$role, "system"), logical(1L))
    sys_text <- paste(vapply(messages[is_sys], `[[`, character(1L), "content"),
                      collapse = "\n")
    body <- list(model = model, max_tokens = max_tokens,
                 temperature = temperature, messages = messages[!is_sys])
    if (nchar(sys_text) > 0L) body$system <- sys_text
    req <- httr2::request("https://api.anthropic.com/v1/messages") |>
      httr2::req_headers("x-api-key"         = api_key,
                         "anthropic-version" = anthropic_version,
                         "Content-Type"      = "application/json") |>
      httr2::req_body_json(body) |>
      httr2::req_retry(
        max_tries    = max_retries,
        is_transient = \(r) httr2::resp_status(r) %in% c(429L, 500L, 529L),
        backoff      = \(i) min(2^i, 60L),
        after        = retry_after_secs
      )
    execute_req(req, \(resp) {
      trimws(httr2::resp_body_json(resp)$content[[1L]]$text)
    }, max_retries = 3L)
  }
  new_provider("Anthropic", model, call_fn)
}

#' Create an AWS Bedrock LLM provider
#'
#' Uses Bedrock's Converse API with AWS SigV4 signing. Supports any model
#' available in your Bedrock account, including Anthropic Claude and Meta
#' Llama families.
#'
#' Credentials are resolved in order: explicit arguments → environment variables
#' (`AWS_ACCESS_KEY_ID`, `AWS_SECRET_ACCESS_KEY`, `AWS_SESSION_TOKEN`,
#' `AWS_DEFAULT_REGION`).
#'
#' Common model IDs:
#' * `"anthropic.claude-haiku-4-5-20251001-v1:0"`
#' * `"anthropic.claude-3-5-sonnet-20241022-v2:0"`
#' * `"meta.llama3-70b-instruct-v1:0"`
#' * `"meta.llama3-8b-instruct-v1:0"`
#' * `"mistral.mistral-7b-instruct-v0:2"`
#'
#' @param model       Bedrock model ID string (see above).
#' @param region      AWS region. Defaults to `AWS_DEFAULT_REGION` env var,
#'   then `"us-east-1"`.
#' @param access_key  AWS access key ID. Defaults to `AWS_ACCESS_KEY_ID`.
#' @param secret_key  AWS secret access key. Defaults to `AWS_SECRET_ACCESS_KEY`.
#' @param session_token Optional STS session token. Defaults to
#'   `AWS_SESSION_TOKEN` (ignored if unset).
#' @param name        Display name for print output and plot labels.
#' @param max_retries Maximum request attempts.
#'
#' @return An `llm_provider` object.
#' @examples
#' \dontrun{
#' p <- llm_provider_bedrock(
#'   model = "anthropic.claude-haiku-4-5-20251001-v1:0"
#' )   # reads AWS_* credentials from env
#' print(p)
#' }
#' @export
llm_provider_bedrock <- function(model,
                                  region        = Sys.getenv("AWS_DEFAULT_REGION",
                                                             unset = "us-east-1"),
                                  access_key    = Sys.getenv("AWS_ACCESS_KEY_ID"),
                                  secret_key    = Sys.getenv("AWS_SECRET_ACCESS_KEY"),
                                  session_token = Sys.getenv("AWS_SESSION_TOKEN"),
                                  name          = "AWS Bedrock",
                                  max_retries   = 5L) {
  force(model); force(region); force(access_key); force(secret_key)
  force(session_token); force(name); force(max_retries)

  tok <- if (nchar(session_token) > 0L) session_token else NULL

  call_fn <- function(messages, max_tokens = 15L, temperature = 0.7) {
    # Converse API: separate system messages from turns
    is_sys   <- vapply(messages, \(m) identical(m$role, "system"), logical(1L))
    sys_text <- paste(vapply(messages[is_sys], `[[`, character(1L), "content"),
                      collapse = "\n")
    turns <- lapply(messages[!is_sys], function(m) {
      list(role    = m$role,
           content = list(list(type = "text", text = m$content)))
    })

    body <- list(
      messages        = turns,
      inferenceConfig = list(maxTokens   = as.integer(max_tokens),
                             temperature = temperature)
    )
    if (nchar(sys_text) > 0L)
      body$system <- list(list(type = "text", text = sys_text))

    url <- sprintf(
      "https://bedrock-runtime.%s.amazonaws.com/model/%s/converse",
      region, utils::URLencode(model, reserved = TRUE)
    )

    req <- httr2::request(url) |>
      httr2::req_headers("Content-Type" = "application/json") |>
      httr2::req_body_json(body) |>
      httr2::req_auth_aws_v4(
        aws_access_key_id     = access_key,
        aws_secret_access_key = secret_key,
        aws_session_token     = tok,
        aws_region            = region,
        aws_service           = "bedrock"
      ) |>
      httr2::req_retry(
        max_tries    = max_retries,
        is_transient = \(r) httr2::resp_status(r) %in% c(429L, 500L, 503L),
        backoff      = \(i) min(2^i, 60L),
        after        = retry_after_secs
      )

    execute_req(req, \(resp) {
      trimws(httr2::resp_body_json(resp)$output$message$content[[1L]]$text)
    }, max_retries = 3L)
  }
  new_provider(name, model, call_fn)
}

#' Call an LLM provider
#'
#' Generic dispatch function. Passes a list of role/content message pairs to
#' the provider's internal call function and returns the text response.
#'
#' @param provider An `llm_provider` object.
#' @param messages List of role/content pairs in OpenAI format, e.g.
#'   `list(list(role = "system", content = "..."), list(role = "user", content = "..."))`.
#' @param max_tokens Token budget for the response.
#' @param temperature Sampling temperature.
#'
#' @return A character string, or `NA_character_` on failure.
#' @examples
#' \dontrun{
#' p    <- llm_provider_openai()
#' msgs <- list(
#'   list(role = "system", content = "Reply with a single integer."),
#'   list(role = "user",   content = "What is 2 + 2?")
#' )
#' llm_call(p, msgs, max_tokens = 5L)
#' }
#' @export
llm_call <- function(provider, messages, max_tokens = 15L, temperature = 0.7) {
  if (!inherits(provider, "llm_provider"))
    stop("`provider` must be an <llm_provider> object.", call. = FALSE)
  provider$call(messages = messages, max_tokens = max_tokens,
                temperature = temperature)
}

test_that("llm_provider_mock returns llm_provider class", {
  p <- llm_provider_mock(mock_strategy_min_id)
  expect_s3_class(p, "llm_provider")
})

test_that("print.llm_provider is silent and returns invisibly", {
  p <- llm_provider_mock(mock_strategy_min_id)
  expect_output(print(p), "Mock")
})

test_that("llm_call errors on non-provider input", {
  expect_snapshot(llm_call("not_a_provider", list()), error = TRUE)
})

test_that("mock_strategy_min_id returns parseable string", {
  p <- llm_provider_mock(mock_strategy_min_id)
  msg <- list(
    list(role = "system", content = "Test."),
    list(role = "user", content = paste0(
      "=== EVENT HISTORY ===\n",
      "  Event 1: Node 1 (1) contacted Node 3 (3).\n",
      "\n=== YOUR TURN ===\n",
      "You are Node 2 (ID: 2).\n",
      "Valid targets: Node 1 (1), Node 3 (3), Node 4 (4)\n",
      "Reply with the integer ID only."
    ))
  )
  resp <- p$call(msg)
  expect_type(resp, "character")
  expect_false(is.na(resp))
  expect_true(as.integer(resp) %in% c(1L, 3L, 4L))
})

test_that("mock_strategy_highest_indegree picks most-targeted node", {
  p <- llm_provider_mock(mock_strategy_highest_indegree)
  # Node 3 receives 2 events, node 1 receives 1 — expect "3"
  msg <- list(
    list(role = "system", content = "Test."),
    list(role = "user", content = paste0(
      "=== EVENT HISTORY ===\n",
      "  Event 1: Node 1 (1) contacted Node 3 (3).\n",
      "  Event 2: Node 2 (2) contacted Node 3 (3).\n",
      "  Event 3: Node 3 (3) contacted Node 1 (1).\n",
      "\n=== YOUR TURN ===\n",
      "You are Node 2 (ID: 2).\n",
      "Valid targets: Node 1 (1), Node 3 (3), Node 4 (4)\n",
      "Reply with the integer ID only."
    ))
  )
  expect_equal(p$call(msg), "3")
})

test_that("mock predict strategy returns two distinct valid integers", {
  p <- llm_provider_mock(mock_strategy_highest_indegree)
  msg <- list(
    list(role = "system", content = "Predict. Reply: SENDER_ID RECEIVER_ID"),
    list(role = "user", content = paste0(
      "=== EVENT HISTORY ===\n",
      "  Event 1: Node 1 (1) contacted Node 3 (3).\n",
      "\n=== PREDICTION TASK ===\n",
      "Actors: Node 1 (1), Node 2 (2), Node 3 (3), Node 4 (4)\n",
      "Reply: SENDER_ID RECEIVER_ID"
    ))
  )
  resp  <- p$call(msg, max_tokens = 25L)
  parts <- suppressWarnings(as.integer(strsplit(trimws(resp), "\\s+")[[1L]]))
  expect_length(parts, 2L)
  expect_false(anyNA(parts))
  expect_false(parts[1L] == parts[2L])
  expect_true(all(parts %in% 1:4))
})

#' @useDynLib llrem, .registration = TRUE
#' @importFrom Rcpp sourceCpp
#' @importFrom stats aggregate coef median model.matrix setNames terms
NULL

# Suppress R CMD check NOTE for ggplot2 aes() variables in compare_rem_models
utils::globalVariables(c("coef", "term", "lower95", "upper95"))

# Conditional logit (discrete-choice REM) -------------------------------------
#
# Ordinal (clogit):
#   log-likelihood = Σ_t  β'x_{i_t j_t}  -  log Σ_{(i,j) ∈ R_t} exp(β'x_{ij})
#   which is McFadden's conditional logit.
#
# Temporal (interval):
#   log-likelihood = Σ_t [ β'x_{i_t j_t}
#                         - log Σ_{(i,j)∈R_t} exp(β'x_{ij})          (choice term)
#                         - Δ_t × Σ_{(i,j)∈R_t} exp(β'x_{ij}) ]     (waiting-time penalty)
#   where Δ_t = t_t − t_{t−1} is the inter-event gap.
#   This is the full Hawkes/REM temporal likelihood; the choice term alone is
#   the ordinal likelihood.  Both share the same risk-set statistics.
#
# Both are maximised via optim(method="BFGS") with analytic gradients.

.rem_clogit_negll <- function(beta, X, strata) {
  eta    <- drop(X %*% beta)               # linear predictor, length = nrow(X)
  # subtract per-stratum max for numerical stability
  s_max  <- tapply(eta, strata, max)[strata]
  eta_c  <- eta - s_max
  exp_e  <- exp(eta_c)
  log_denom <- log(tapply(exp_e, strata, sum))[strata]
  lp     <- eta_c - log_denom              # log P(choose this row | stratum)
  ll     <- sum(lp[X[, 1L] == X[, 1L]])   # dummy; will use occurred mask below
  ll
}

# Fit conditional logit REM via BFGS.
# rs      : risk-set data frame with column 'occurred' and covariate columns
# formula : one-sided ~ covariate formula (no Surv(), no strata())
# Returns a list mimicking enough of coxph to be useful: coef, vcov, loglik
.fit_clogit <- function(rs, formula) {
  covars <- all.vars(formula)
  X      <- as.matrix(rs[, covars, drop = FALSE])
  y      <- rs$occurred
  g      <- rs$time2   # stratum = event time (unique per event)

  # negative log-likelihood
  negll <- function(beta) {
    eta   <- drop(X %*% beta)
    s_max <- tapply(eta, g, max)[as.character(g)]
    ec    <- exp(eta - s_max)
    ld    <- log(tapply(ec, g, sum))[as.character(g)]
    -sum((eta - s_max - ld)[y == 1L])
  }

  # analytic gradient
  grad <- function(beta) {
    eta   <- drop(X %*% beta)
    s_max <- tapply(eta, g, max)[as.character(g)]
    ec    <- exp(eta - s_max)
    denom <- tapply(ec, g, sum)[as.character(g)]
    p     <- ec / denom                   # P(row | stratum)
    # gradient = Σ_t [ x_{chosen_t} - E[x | stratum t] ]
    # = X'(y - p)  (where y is the 0/1 occurred indicator)
    -drop(t(X) %*% (y - p))
  }

  beta0 <- rep(0, ncol(X))
  fit   <- stats::optim(beta0, negll, grad, method = "BFGS",
                        control = list(maxit = 500L),
                        hessian = TRUE)

  coef <- fit$par
  names(coef) <- covars

  # vcov from observed Fisher information
  vcov <- tryCatch(solve(fit$hessian), error = function(e) {
    matrix(NA_real_, ncol(X), ncol(X),
           dimnames = list(covars, covars))
  })
  dimnames(vcov) <- list(covars, covars)

  structure(
    list(coef   = coef,
         vcov   = vcov,
         loglik = c(negll(rep(0, ncol(X))), -fit$value),
         converged = fit$convergence == 0L,
         method = "clogit"),
    class = "rem_clogit"
  )
}

# Fit temporal REM via BFGS.
#
# The temporal (interval-censored) log-likelihood is:
#   LL = Σ_{t=2}^{m} [ β'x_{focal,t}  −  Δ_t × Σ_{ij} exp(β'x_{ij,t}) ]
#        − Δ_{m+1} × Σ_{ij} exp(β'x_{ij,m})       ← right-censoring at end
#
# where Δ_t = t_t − t_{t−1} and the outer sum is over all n*(n-1) dyads.
# Unlike the ordinal likelihood there is no log-normaliser; the rate of the
# focal dyad enters directly, not as a conditional probability.
#
# With risk-set sampling the population total rate is estimated as:
#   Λ_t ≈ (n_dyads / ns_t) × Σ_{ij ∈ sample_t} exp(β'x_{ij,t})
# This is unbiased when exp(η) is roughly uniform across dyads; it degrades
# for very skewed rate distributions (a known limitation of rate-model sampling).
#
# rs      : risk-set data frame (remstats_generate output + add_survival_cols)
# formula : one-sided ~ covariate formula
# deltas  : numeric vector length m, delta[t] = t_t − t_{t−1}, delta[1] = t_1
# n_dyads : n*(n-1); used to scale the sampled total rate back to population
.fit_temporal <- function(rs, formula, deltas, n_dyads) {
  covars <- all.vars(formula)
  X      <- as.matrix(rs[, covars, drop = FALSE])
  y      <- rs$occurred
  g      <- rs$time2   # stratum id = event time (unique per event, events 2..m)

  strata_ord <- sort(unique(g))               # length = m - 1
  # deltas for strata 2..m are deltas[2..m]; index via position in strata_ord
  # strata_ord[k] is the k-th unique event time; its gap is deltas[k+1]
  # (deltas[1] = t_1, not used in the sum — event 1 has no risk set in pe)
  delta_u   <- deltas[seq(2L, length(deltas))]   # length = m - 1
  delta_map <- delta_u[match(g, strata_ord)]      # per risk-set row

  # sampling scale: n_dyads / ns_t per stratum
  ns_tab    <- table(g)
  scale_u   <- n_dyads / as.numeric(ns_tab)       # length = m - 1
  scale_map <- n_dyads / as.numeric(ns_tab[as.character(g)])

  negll <- function(beta) {
    eta <- drop(X %*% beta)
    # focal log-rates: Σ_t β'x_{focal,t}
    log_rate_focal <- sum(eta[y == 1L])
    # total rate per stratum: Λ_t ≈ scale_u * Σ_sample exp(η)
    # use log-sum-exp for stability
    s_max  <- tapply(eta, g, max)
    ec     <- exp(eta - s_max[as.character(g)])
    sum_ec <- as.numeric(tapply(ec, g, sum))       # length = m - 1
    Lambda <- scale_u * sum_ec * exp(s_max)
    # negative log-likelihood
    -(log_rate_focal - sum(delta_u * Lambda))
  }

  grad <- function(beta) {
    eta <- drop(X %*% beta)
    # focal contribution: -x_{focal,t} per event
    g_focal <- -colSums(X[y == 1L, , drop = FALSE])
    # penalty gradient: Σ_t Δ_t × scale_t × Σ_{ij∈sample_t} exp(η_ij) × x_ij
    ec_abs  <- exp(eta)
    g_pen   <- drop(t(X) %*% (delta_map * scale_map * ec_abs))
    g_focal + g_pen
  }

  beta0 <- rep(0, ncol(X))
  fit   <- stats::optim(beta0, negll, grad, method = "BFGS",
                        control = list(maxit = 500L),
                        hessian = TRUE)

  coef <- fit$par
  names(coef) <- covars

  vcov <- tryCatch(solve(fit$hessian), error = function(e) {
    matrix(NA_real_, ncol(X), ncol(X), dimnames = list(covars, covars))
  })
  dimnames(vcov) <- list(covars, covars)

  structure(
    list(coef      = coef,
         vcov      = vcov,
         loglik    = c(-negll(rep(0, ncol(X))), -fit$value),
         converged = fit$convergence == 0L,
         method    = "temporal"),
    class = "rem_clogit"
  )
}

#' @export
coef.rem_clogit <- function(object, ...) object$coef

#' @export
print.rem_clogit <- function(x, ...) {
  se  <- sqrt(diag(x$vcov))
  z   <- x$coef / se
  tab <- data.frame(coef = x$coef, `se(coef)` = se, z = z,
                    `Pr(>|z|)` = 2 * stats::pnorm(-abs(z)),
                    check.names = FALSE)
  label <- if (identical(x$method, "temporal")) "REM temporal" else "REM conditional logit"
  cat(label, "\n\n")
  print(round(tab, 4L))
  if (!x$converged) cat("\n[Warning: optimizer did not converge]\n")
  invisible(x)
}

# covariate formula from a Surv formula (strips Surv() and strata())
.surv_to_covar_formula <- function(formula) {
  tt  <- terms(formula, specials = c("strata", "survival::strata"))
  labs <- attr(tt, "term.labels")
  # drop any term whose label contains "strata("
  labs <- labs[!grepl("strata\\(", labs)]
  stats::as.formula(paste("~", paste(labs, collapse = " + ")))
}

#' Compute sufficient-statistic means for a sequence without fitting a REM
#'
#' Builds covariate values only for the realized events in `edgelist`
#' (n_nonevents = 0L) and returns their column means.  Much faster than
#' [fit_rem()] since no Cox model is fitted and the risk set contains exactly
#' one row per event.  Use to measure convergence in sufficient-statistic space
#' rather than coefficient space.
#'
#' @param edgelist    Prepared edgelist (output of [prepare_edgelist()]).
#' @param rem_formula Survival formula — used to identify covariate columns.
#' @param base_riskset Precomputed base risk set; constructed from `edgelist`
#'   if `NULL`.
#' @param covariates Optional exogenous covariate list from [rem_covariates()].
#'
#' @return Named numeric vector: mean of each REM covariate across events.
#' @examples
#' data(edgelist)
#' el <- prepare_edgelist(edgelist)
#' compute_seq_stat_means(
#'   el[1:100, ],
#'   survival::Surv(time1, occurred) ~ Repetition + RecRecencyLog +
#'     survival::strata(time2)
#' )
#' @export
compute_seq_stat_means <- function(edgelist, rem_formula, base_riskset = NULL,
                                    covariates = NULL) {
  all_nodes    <- get_nodes(edgelist, NULL)
  base_riskset <- get_base_riskset(base_riskset, all_nodes)

  pe <- remstats_generate(edgelist,
                           n            = length(all_nodes),
                           n_nonevents  = 0L,
                           base_riskset = base_riskset,
                           covariates   = covariates)
  pe <- add_survival_cols(pe)

  ss <- make_suffstats(rem_formula, pe, drop_strata = TRUE)
  colMeans(ss, na.rm = TRUE)
}

#' Fit a Relational Event Model to an edgelist
#'
#' Builds risk-set statistics for each event in the sequence, adds survival
#' model timing columns (`time1`, `time2`), and fits a Cox proportional hazards
#' model. All covariate computation is handled internally.
#'
#' @param edgelist Data frame with columns `time`, `sender`, `receiver`,
#'   `eveid`. Use [prepare_edgelist()] to create this from a raw matrix.
#' @param formula A `survival::Surv` formula. Must reference `time1` and
#'   `time2` (via `strata()`).
#'
#'   **Available covariates**
#'
#'   \tabular{llll}{
#'     \strong{Name} \tab \strong{Definition} \tab \strong{relevent analogue} \tab \strong{Notes} \cr
#'     \emph{Dyad activity} \tab \tab \tab \cr
#'     `Repetition` \tab dyad_occ / sender_outDeg \tab `FrPSndSnd` \tab Same quantity; llrem denom = n-1 (fixed), relevent = event index \cr
#'     `SndInitiation` \tab I(dyad i→j has ever fired) \tab — \tab llrem only; used as partition variable in [fit_rem_sep()] \cr
#'     `DyadRecencyLog` \tab -log1p(time since last i→j) \tab `RSndSnd` \tab Same direction; relevent uses recency rank, not log-time \cr
#'     `WeightedRep` \tab w_ij / max(w_ij in risk set) \tab — \tab llrem only; same-dyad count normalised by per-sender max \cr
#'     \emph{Reciprocity} \tab \tab \tab \cr
#'     `RRecInitiation` \tab I(reverse dyad j→i has ever fired) \tab `PSAB-BA` (one-step) \tab relevent PSAB-BA resets after each non-BA event; RRecInitiation is permanent \cr
#'     `RecRecencyLog` \tab -log1p(time since last j→i) \tab `RRecSnd` \tab Both measure reverse-dyad recency; relevent uses rank, llrem uses log-time \cr
#'     `RecRecencyRank` \tab 1 / rank(j in i's incoming recency list) \tab `RRecSnd` \tab Exact relevent analogue; rank 1 = most recent sender to i \cr
#'     `WeightedRec` \tab min(w_ij, w_ji) / max(w_ij, w_ji) \tab — \tab Squartini et al. (2013) weighted reciprocity; 0 if both zero \cr
#'     `FrRecSnd` \tab w_ji / sender_inDeg \tab `FrRecSnd` \tab Fraction of sender's in-events that are from receiver \cr
#'     \emph{Degree} \tab \tab \tab \cr
#'     `OutDegSnd` \tab sender out-degree / (n-1) \tab `NODSnd` \tab relevent divides by running event count \cr
#'     `OutDegRec` \tab receiver out-degree / (n-1) \tab `NODRec` \tab same normalisation difference \cr
#'     `InDegSnd` \tab sender in-degree / (n-1) \tab `NIDSnd` \tab same normalisation difference \cr
#'     `InDegRec` \tab receiver in-degree / (n-1) \tab `NIDRec` \tab same normalisation difference \cr
#'     `TotDegSnd` \tab (OutDegSnd + InDegSnd) / 2 \tab `NTDegSnd` \tab both are total-degree averages \cr
#'     `TotDegRec` \tab (OutDegRec + InDegRec) / 2 \tab `NTDegRec` \tab both are total-degree averages \cr
#'     \emph{Triadic closure} \tab \tab \tab \cr
#'     `OTP` \tab out-two-path count \tab `OTPSnd` \tab raw count; use `NOTPSnd` for normalised \cr
#'     `ITP` \tab in-two-path count \tab `ITPSnd` \tab raw count; use `NITPSnd` for normalised \cr
#'     `OSP` \tab out-shared-partner count \tab `OSPSnd` \tab raw count; use `NOSPSnd` for normalised \cr
#'     `ISP` \tab in-shared-partner count \tab `ISPSnd` \tab raw count; use `NISPSnd` for normalised \cr
#'     `NOTPSnd` \tab OTP / max(OTP in risk set) \tab — \tab llrem only (normalised OTP) \cr
#'     `NITPSnd` \tab ITP / max(ITP in risk set) \tab — \tab llrem only (normalised ITP) \cr
#'     `NOSPSnd` \tab OSP / max(OSP in risk set) \tab — \tab llrem only (normalised OSP) \cr
#'     `NISPSnd` \tab ISP / max(ISP in risk set) \tab — \tab llrem only (normalised ISP) \cr
#'     `NTriadic` \tab (OTP + ITP + OSP + ISP) / max(OTP + ITP + OSP + ISP) \tab — \tab llrem only; omnibus triadic score; ranges 0--1 \cr
#'     \emph{Imitation} \tab \tab \tab \cr
#'     `Imitation` \tab (recInDg - w_ij) / max per sender \tab — \tab Times j targeted by actors other than i, normalised; ranges 0--1 \cr
#'     `PreImitation` \tab recInDg - w_ij (raw, unnormalised) \tab — \tab Raw count underlying Imitation \cr
#'     \emph{P-shifts} \tab \tab \tab \cr
#'     `IOviolation` \tab PSAB_XA + PSAB_XB + PSAB_XY (any turn-usurp) \tab `PSAB-XA`/`PSAB-XB`/`PSAB-XY` (separate) \tab llrem collapses the three usurping types \cr
#'     `PSAB_BA` \tab I(last event was j→i, this is i→j) \tab `PSAB-BA` \tab exact match \cr
#'     `PSAB_BY` \tab I(last event was j→i, this is i→y, y!=j) \tab `PSAB-BY` \tab exact match \cr
#'     `PSAB_AY` \tab I(last event was i→j, this is i→y, y!=j) \tab `PSAB-AY` \tab exact match \cr
#'     `PSAB_XA` \tab I(last event was x→i, this is j→i) \tab `PSAB-XA` \tab exact match \cr
#'     `PSAB_XB` \tab I(last event was x→j, this is i→j) \tab `PSAB-XB` \tab exact match \cr
#'     `PSAB_XY` \tab I(last event was x→y, this is i→j; x!=i, y!=j) \tab `PSAB-XY` \tab exact match \cr
#'   }
#' @param n Number of nodes. Inferred from `edgelist` if `NULL`.
#' @param n_nonevents Sampled non-event risk-set size per event. `NULL` uses
#'   the full risk set (slow for large networks; 200 is a sensible default).
#' @param base_riskset Precomputed base risk-set data frame. Constructed from
#'   `n` if `NULL`. Pass a cached object when calling `fit_rem()` and
#'   [run_llm_rem()] in the same session to avoid redundant work.
#' @param cl Optional PSOCK cluster from `parallel::makeCluster()` for
#'   parallelised risk-set computation.
#' @param method Estimation method: `"cox"` (default) fits a Cox proportional
#'   hazards model via [survival::coxph()] with one stratum per event.
#'   `"clogit"` fits a conditional logit (discrete-choice/ordinal) model via
#'   BFGS; substantially faster than `"cox"` for large sequences and
#'   asymptotically equivalent under the ordinal likelihood.
#'   `"temporal"` fits the full interval-censored REM likelihood, which adds a
#'   waiting-time penalty `Delta_t * Sigma exp(eta)` to each event's contribution.
#'   Requires that `edgelist$time` contains real inter-event gaps, not just
#'   ranks; coefficients are no longer scale-free.
#'
#'   Available covariates: see `@param formula` above.
#'
#'   Note on naming vs **relevent**: `OutDegSnd`/`OutDegRec`/`InDegSnd`/`InDegRec`
#'   scale by `(n-1)` (fixed maximum degree), unlike relevent's `NODSnd`/`NIDRec`
#'   which scale by the running event count.  `DyadRecencyLog` is
#'   `-log1p(clock-time of last i→j event)`; relevent's `RSndSnd` is a recency
#'   rank -- both measure same-dyad recency but are not numerically comparable.
#'   `RecRecencyLog` (`-log1p(time since last j→i)`) and `RecRecencyRank`
#'   (`1/rank` of j in i's incoming recency list) are both analogues of
#'   relevent's `RRecSnd`; `RecRecencyRank` is the exact functional equivalent.
#'
#' @param covariates Optional exogenous covariate list created by
#'   [rem_covariates()]. `NULL` uses only the built-in endogenous statistics.
#' @return For `method = "cox"`: a `coxph` object.  For `method = "clogit"` or
#'   `"temporal"`: a `rem_clogit` object with elements `coef`, `vcov`,
#'   `loglik`, and `converged`.
#' @examples
#' data(edgelist)
#' el <- prepare_edgelist(edgelist)
#' fit <- fit_rem(
#'   el[1:100, ],
#'   survival::Surv(time1, occurred) ~ Repetition + RecRecencyLog +
#'     survival::strata(time2),
#'   n_nonevents = 20L, method = "clogit"
#' )
#' print(fit)
#' @export
fit_rem <- function(edgelist,
                    formula,
                    n            = NULL,
                    n_nonevents  = 200L,
                    base_riskset = NULL,
                    cl           = NULL,
                    method       = c("cox", "clogit", "temporal"),
                    covariates   = NULL) {
  method       <- match.arg(method)
  all_nodes    <- get_nodes(edgelist, n)
  base_riskset <- get_base_riskset(base_riskset, all_nodes)
  pe           <- remstats_generate(edgelist,
                                     n            = length(all_nodes),
                                     n_nonevents  = n_nonevents,
                                     cl           = cl,
                                     base_riskset = base_riskset,
                                     covariates   = covariates)
  pe <- add_survival_cols(pe)

  if (method %in% c("clogit", "temporal")) {
    covar_formula <- .surv_to_covar_formula(formula)
    if (method == "clogit")
      return(.fit_clogit(pe, covar_formula))
    # inter-event gaps in the same order as sort(unique(pe$time2))
    times   <- sort(unique(edgelist$time))
    deltas  <- c(times[1], diff(times))   # delta[1] = t_1 (first gap from 0)
    n_dyads <- length(all_nodes) * (length(all_nodes) - 1L)
    return(.fit_temporal(pe, covar_formula, deltas, n_dyads))
  }
  survival::coxph(formula, data = pe)
}

#' Generate an LLM event sequence and fit a REM
#'
#' At each time step *t* (starting from 2), `query_fn` is called with the
#' growing LLM-generated history, the empirical event at *t*, and the full
#' node set. The returned sender/receiver pair is appended to the sequence.
#' The sequence is seeded with the first empirical event so the LLM starts
#' with at least one data point. After generation, risk-set statistics are
#' built and a REM is fitted.
#'
#' @param edgelist Prepared edgelist (output of [prepare_edgelist()]).
#' @param formula Survival formula passed to [fit_rem()].
#' @param query_fn A closure produced by [make_query_roleplay_sender()],
#'   [make_query_predict()], or [make_query_roleplay_random()], or any function
#'   with the signature:
#'   `f(past_events, empirical_event, all_nodes) -> c(sender, receiver)`.
#' @param n_events Number of events to generate. `NULL` uses the full edgelist.
#' @param n Number of nodes. Inferred if `NULL`.
#' @param n_nonevents Sampled risk-set size passed to [fit_rem()].
#' @param base_riskset Precomputed base risk set.
#' @param cl Optional PSOCK cluster for parallelised risk-set computation.
#' @param seed RNG seed for random fallbacks and random-node sampling.
#' @param checkpoint_path File path for incremental saves every 10 events.
#'   `NULL` disables checkpointing.
#' @param use_empirical_history If `FALSE` (default), the LLM is shown its own
#'   previously generated events as history, so any divergence from reality
#'   compounds over time. If `TRUE`, the LLM is shown the empirical event
#'   history at each step, isolating the effect of the generation strategy from
#'   the effect of accumulated sequence drift.
#' @param rate_limit_secs Seconds to pause between LLM calls.
#' @param fit If `TRUE` (default), fits a REM on the generated sequence before
#'   returning. Set to `FALSE` to skip model fitting and return only the
#'   sequence and raw responses — useful when running many sequences and
#'   deferring all model fitting to a separate step.
#' @param method Passed to [fit_rem()]: `"cox"` (default), `"clogit"`, or
#'   `"temporal"`.
#' @param covariates Optional exogenous covariate list passed to [fit_rem()].
#'
#' @return A named list with elements:
#'   - `sequence`: generated event sequence as a data frame.
#'   - `responses`: raw LLM response text and parsed sender/receiver per event.
#'   - `rem`: fitted model (only present when `fit = TRUE`).
#' @examples
#' \dontrun{
#' data(edgelist)
#' el  <- prepare_edgelist(edgelist)
#' cfg <- rem_cfg(
#'   provider    = llm_provider_openai(),
#'   context     = "These are militarized conflict events between countries.",
#'   action_verb = "attacked",
#'   actor_noun  = "Country"
#' )
#' res <- run_llm_rem(
#'   el[1:50, ],
#'   survival::Surv(time1, occurred) ~ Repetition + RecRecencyLog +
#'     survival::strata(time2),
#'   query_fn    = make_query_roleplay_sender(cfg),
#'   n_nonevents = 50L, method = "clogit"
#' )
#' print(res$rem)
#' }
#' @export
run_llm_rem <- function(edgelist,
                         formula,
                         query_fn,
                         n_events              = NULL,
                         n                     = NULL,
                         n_nonevents           = 200L,
                         base_riskset          = NULL,
                         cl                    = NULL,
                         seed                  = 8214L,
                         checkpoint_path       = NULL,
                         use_empirical_history = FALSE,
                         rate_limit_secs       = 0.3,
                         fit                   = TRUE,
                         method                = c("cox", "clogit", "temporal"),
                         covariates            = NULL) {
  method <- match.arg(method)
  all_nodes    <- get_nodes(edgelist, n)
  base_riskset <- get_base_riskset(base_riskset, all_nodes)

  n_ev <- if (is.null(n_events)) nrow(edgelist) else min(n_events, nrow(edgelist))
  emp  <- edgelist[seq_len(n_ev), ]

  set.seed(seed)
  llm_rows <- list()
  start_t  <- 2L

  if (!is.null(checkpoint_path) && file.exists(checkpoint_path)) {
    env <- new.env(parent = emptyenv())
    load(checkpoint_path, envir = env)
    llm_rows <- env$llm_rows
    start_t  <- length(llm_rows) + 2L
    message("[llrem] Resuming from checkpoint at event ", start_t)
  }

  llm_seq <- emp[1L, c("time", "sender", "receiver", "eveid")]

  if (start_t > n_ev) {
    message("[llrem] Checkpoint is complete; skipping generation.")
  } else {
    pb <- utils::txtProgressBar(min = start_t, max = n_ev, style = 3)

    for (t in start_t:n_ev) {
      history <- if (use_empirical_history) emp[seq_len(t - 1L), ] else llm_seq
      result <- query_fn(
        past_events     = history,
        empirical_event = emp[t, ],
        all_nodes       = all_nodes
      )
      new_row <- data.frame(
        time     = emp$time[t],
        sender   = as.integer(result[["sender"]]),
        receiver = as.integer(result[["receiver"]]),
        eveid    = paste0(result[["sender"]], "_", result[["receiver"]]),
        stringsAsFactors = FALSE
      )
      llm_rows[[t - 1L]] <- new_row
      llm_seq <- rbind(llm_seq, new_row[, c("time", "sender", "receiver", "eveid")])

      if (!is.null(checkpoint_path) && (t %% 10L == 0L || t == n_ev))
        save(llm_rows, file = checkpoint_path)

      utils::setTxtProgressBar(pb, t)
      Sys.sleep(rate_limit_secs)
    }
    close(pb)
  }

  sequence <- rbind(emp[1L, c("time", "sender", "receiver", "eveid")],
                    do.call(rbind, llm_rows))
  row.names(sequence) <- NULL

  out <- list(sequence  = sequence,
              responses = do.call(rbind, llm_rows))

  if (fit) {
    message("\n[llrem] Fitting REM on generated sequence...")
    out$rem <- fit_rem(sequence, formula,
                       n = length(all_nodes), n_nonevents = n_nonevents,
                       base_riskset = base_riskset, cl = cl, method = method,
                       covariates = covariates)
  }

  out
}

# ---------------------------------------------------------------------------
# Separable REM
# ---------------------------------------------------------------------------

#' Fit a Separable Relational Event Model
#'
#' Partitions the risk set by prior contact status (`SndInitiation`) and fits
#' two independent sub-models: one for **initiation** events (sender has never
#' previously contacted receiver; `SndInitiation == 0`) and one for
#' **continuation** events (sender has contacted receiver before;
#' `SndInitiation == 1`).  The approach is analogous to separable TERGMs
#' (Krivitsky and Handcock, 2014): because the two subsets are disjoint, their
#' parameters are conditionally independent and can be estimated separately via
#' standard Cox partial likelihood on each partition.
#'
#' The combined log-likelihood is `loglik_init + loglik_cont`, which is directly
#' comparable to a pooled five-parameter model with the same degrees of freedom
#' (see paper for details).
#'
#' @param edgelist Data frame with columns `time`, `sender`, `receiver`,
#'   `eveid`. Use [prepare_edgelist()] to create this from a raw matrix.
#' @param formula_init Survival formula for the initiation sub-model
#'   (events where `SndInitiation == 0`).  Typical terms: `RRecInitiation`,
#'   `NOTPSnd`, `NITPSnd`.  Must include `survival::strata(time2)`.
#' @param formula_cont Survival formula for the continuation sub-model
#'   (events where `SndInitiation == 1`).  Typical terms: `Repetition`,
#'   `RecRecencyLog`.  Must include `survival::strata(time2)`.
#' @param n Number of nodes. Inferred from `edgelist` if `NULL`.
#' @param n_nonevents Sampled non-event risk-set size per event. `NULL` uses
#'   the full risk set. Passed to [fit_rem()] internally.
#' @param base_riskset Precomputed base risk-set data frame.
#' @param cl Optional PSOCK cluster for parallelised risk-set computation.
#' @param method Estimation method: `"cox"` (default), `"clogit"`, or
#'   `"temporal"`. Applied independently to each sub-model.
#' @param covariates Optional exogenous covariate list from [rem_covariates()].
#' @param min_events Minimum number of focal events required in each partition
#'   to attempt fitting. Partitions below this threshold return `NULL` for
#'   that sub-model with a warning. Default 2.
#'
#' @return A `rem_sep` object (a named list) with elements:
#'   \describe{
#'     \item{`init`}{Fitted initiation sub-model (same class as [fit_rem()]
#'       output for the chosen `method`), or `NULL` if too few events.}
#'     \item{`cont`}{Fitted continuation sub-model, or `NULL`.}
#'     \item{`loglik`}{Named numeric vector: `init`, `cont`, `sep`
#'       (= init + cont), and `null` (= init_null + cont_null).}
#'     \item{`n_init`}{Number of initiation events in the sequence.}
#'     \item{`n_cont`}{Number of continuation events.}
#'     \item{`init_fraction`}{Proportion of events that are initiations.}
#'     \item{`method`}{The estimation method used.}
#'   }
#' @examples
#' data(edgelist)
#' el <- prepare_edgelist(edgelist)
#' fit <- fit_rem_sep(
#'   el[1:200, ],
#'   formula_init = survival::Surv(time1, occurred) ~
#'     RRecInitiation + survival::strata(time2),
#'   formula_cont = survival::Surv(time1, occurred) ~
#'     Repetition + RecRecencyLog + survival::strata(time2),
#'   n_nonevents = 20L, method = "clogit"
#' )
#' print(fit)
#' @export
fit_rem_sep <- function(edgelist,
                        formula_init,
                        formula_cont,
                        n            = NULL,
                        n_nonevents  = 200L,
                        base_riskset = NULL,
                        cl           = NULL,
                        method       = c("cox", "clogit", "temporal"),
                        covariates   = NULL,
                        min_events   = 2L) {
  method       <- match.arg(method)
  all_nodes    <- get_nodes(edgelist, n)
  base_riskset <- get_base_riskset(base_riskset, all_nodes)

  pe <- remstats_generate(edgelist,
                           n            = length(all_nodes),
                           n_nonevents  = n_nonevents,
                           cl           = cl,
                           base_riskset = base_riskset,
                           covariates   = covariates)
  pe <- add_survival_cols(pe)

  # Partition ----------------------------------------------------------------
  # A stratum (event time) contains exactly one focal row; that row belongs
  # entirely to one partition.  Strata that fall in the other partition must
  # be dropped so each sub-model's partial likelihood is well-formed.
  pe_init <- pe[pe$SndInitiation == 0, ]
  pe_cont <- pe[pe$SndInitiation == 1, ]

  # Keep only strata that contain at least one focal event in the partition
  keep_strata <- function(sub) {
    valid <- unique(sub$time2[sub$occurred == 1L])
    sub[sub$time2 %in% valid, ]
  }
  pe_init <- keep_strata(pe_init)
  pe_cont <- keep_strata(pe_cont)

  n_init <- sum(pe$occurred[pe$SndInitiation == 0])
  n_cont <- sum(pe$occurred[pe$SndInitiation == 1])

  # Fit helper: routes to the right estimator --------------------------------
  .fit_partition <- function(pe_sub, formula, label) {
    n_ev <- sum(pe_sub$occurred)
    if (n_ev < min_events) {
      warning(sprintf("[fit_rem_sep] %s sub-model skipped: only %d focal event(s).",
                      label, n_ev))
      return(NULL)
    }
    if (method == "cox")
      return(survival::coxph(formula, data = pe_sub))

    covar_formula <- .surv_to_covar_formula(formula)
    if (method == "clogit")
      return(.fit_clogit(pe_sub, covar_formula))

    # temporal: deltas from the full edgelist time axis (gaps are sequence-wide)
    times   <- sort(unique(edgelist$time))
    deltas  <- c(times[1L], diff(times))
    n_dyads <- length(all_nodes) * (length(all_nodes) - 1L)
    .fit_temporal(pe_sub, covar_formula, deltas, n_dyads)
  }

  fit_init <- .fit_partition(pe_init, formula_init, "initiation")
  fit_cont <- .fit_partition(pe_cont, formula_cont, "continuation")

  # Log-likelihoods ----------------------------------------------------------
  .ll <- function(fit) {
    if (is.null(fit)) return(c(null = NA_real_, fit = NA_real_))
    if (inherits(fit, "coxph"))
      return(c(null = fit$loglik[1L], fit = fit$loglik[2L]))
    c(null = fit$loglik[1L], fit = fit$loglik[2L])
  }
  ll_i <- .ll(fit_init)
  ll_c <- .ll(fit_cont)

  loglik <- c(
    init = unname(ll_i["fit"]),
    cont = unname(ll_c["fit"]),
    sep  = unname(ll_i["fit"]) + unname(ll_c["fit"]),
    null = unname(ll_i["null"]) + unname(ll_c["null"])
  )

  structure(
    list(
      init          = fit_init,
      cont          = fit_cont,
      loglik        = loglik,
      n_init        = n_init,
      n_cont        = n_cont,
      init_fraction = n_init / (n_init + n_cont),
      method        = method
    ),
    class = "rem_sep"
  )
}

#' @export
print.rem_sep <- function(x, ...) {
  cat(sprintf("Separable REM (%s)\n", x$method))
  cat(sprintf("  Events: %d initiation (%.1f%%)  |  %d continuation\n",
              x$n_init, 100 * x$init_fraction, x$n_cont))
  cat(sprintf("  Log-lik: init = %.1f  cont = %.1f  sep = %.1f\n\n",
              x$loglik["init"], x$loglik["cont"], x$loglik["sep"]))

  if (!is.null(x$init)) {
    cat("-- Initiation sub-model --\n")
    print(x$init, ...)
  } else {
    cat("-- Initiation sub-model: NULL (too few events) --\n")
  }
  cat("\n")
  if (!is.null(x$cont)) {
    cat("-- Continuation sub-model --\n")
    print(x$cont, ...)
  } else {
    cat("-- Continuation sub-model: NULL (too few events) --\n")
  }
  invisible(x)
}

make_two_models <- function() {
  m1 <- fit_rem(test_edgelist, test_formula,
                n_nonevents = NULL, base_riskset = test_riskset)
  res <- run_llm_rem(
    test_edgelist, test_formula,
    query_fn     = make_query_roleplay_sender(mock_cfg),
    n_nonevents  = NULL, base_riskset = test_riskset, seed = 1L
  )
  list("Empirical" = m1, "LLM" = res$rem)
}

test_that("compare_rem_models returns a ggplot invisibly", {
  models <- make_two_models()
  p <- suppressWarnings(compare_rem_models(models, title = "Test"))
  expect_s3_class(p, "gg")
})

test_that("compare_rem_models data has one row per term per model", {
  models   <- make_two_models()
  p        <- suppressWarnings(compare_rem_models(models))
  n_terms  <- length(names(stats::coef(models[[1L]])))
  expect_equal(nrow(p$data), n_terms * length(models))
})

test_that("term_order sets factor levels correctly", {
  models <- make_two_models()
  ord    <- c("NTriadic", "IOviolation", "RecRecencyLog",
              "SndInitiation", "Repetition", "Imitation")
  p      <- suppressWarnings(compare_rem_models(models, term_order = ord))
  expect_equal(levels(p$data$term), ord)
})

test_that("custom palette is applied to rendered point colours", {
  models  <- make_two_models()
  palette <- c("Empirical" = "#FF0000", "LLM" = "#0000FF")
  p       <- suppressWarnings(compare_rem_models(models, palette = palette))
  built   <- suppressWarnings(ggplot2::ggplot_build(p))
  colours <- toupper(unique(built$data[[3L]]$colour))
  expect_true(all(c("#FF0000", "#0000FF") %in% colours))
})

test_that("compare_rem_models errors on an unnamed model list", {
  models <- make_two_models()
  expect_snapshot(compare_rem_models(unname(models)), error = TRUE)
})

# Prompt utilities (internal) -------------------------------------------------

node_label <- function(id, cfg) {
  if (!is.null(cfg$node_labels)) {
    lbl <- cfg$node_labels[as.character(id)]
    if (!is.na(lbl)) return(lbl)
  }
  paste(cfg$actor_noun, id)
}

format_history <- function(past_events, cfg) {
  if (nrow(past_events) == 0L) return("(No prior events.)")
  window <- cfg$history_window
  if (!is.null(window) && nrow(past_events) > window)
    past_events <- utils::tail(past_events, window)
  lines <- paste0(
    "  Event ", past_events$time, ": ",
    node_label(past_events$sender,   cfg), " (", past_events$sender,   ") ",
    cfg$action_verb, " ",
    node_label(past_events$receiver, cfg), " (", past_events$receiver, ")."
  )
  paste(lines, collapse = "\n")
}

make_messages <- function(system_msg, user_msg) {
  list(list(role = "system", content = system_msg),
       list(role = "user",   content = user_msg))
}

first_valid_int <- function(text, valid_set) {
  if (length(text) == 0L || is.na(text) || !nzchar(text)) return(NA_integer_)
  nums <- suppressWarnings(
    as.integer(regmatches(text, gregexpr("-?[0-9]+", text))[[1L]])
  )
  nums <- nums[nums %in% valid_set]
  if (length(nums) > 0L) nums[1L] else NA_integer_
}

label_match <- function(text, valid_set, cfg) {
  for (v in valid_set)
    if (grepl(node_label(v, cfg), text, ignore.case = TRUE)) return(v)
  NA_integer_
}

# Query factories -------------------------------------------------------------

#' Query factory: LLM roleplays as the empirical sender
#'
#' Returns a `query_fn` closure in which the LLM is told it IS the empirical
#' sender at time *t* and asked to choose a receiver. The generated sequence
#' preserves the empirical sender ordering but uses LLM-selected receivers.
#'
#' @param cfg A `rem_cfg` object.
#'
#' @return A function with signature
#'   `f(past_events, empirical_event, all_nodes) -> c(sender, receiver)`.
#' @examples
#' \dontrun{
#' cfg <- rem_cfg(
#'   provider    = llm_provider_openai(),
#'   context     = "Conflict events between states.",
#'   action_verb = "attacked",
#'   actor_noun  = "Country"
#' )
#' query_fn <- make_query_roleplay_sender(cfg)
#' }
#' @export
make_query_roleplay_sender <- function(cfg) {
  force(cfg)
  function(past_events, empirical_event, all_nodes) {
    sender_id     <- empirical_event$sender
    valid_targets <- setdiff(all_nodes, sender_id)
    target_str    <- paste(
      paste0(node_label(valid_targets, cfg), " (", valid_targets, ")"),
      collapse = ", "
    )
    msgs <- make_messages(
      paste0(cfg$context,
             "\n\nYou are roleplaying as one actor in this network. ",
             "Given the event history, decide who you will target next. ",
             "Reply with ONLY the integer ID of your target -- no other text."),
      paste0("=== EVENT HISTORY ===\n", format_history(past_events, cfg),
             "\n\n=== YOUR TURN ===\n",
             "You are ", node_label(sender_id, cfg), " (ID: ", sender_id, ").\n",
             "Valid targets: ", target_str, "\n",
             "Which actor do you target? Reply with the integer ID only.")
    )
    raw  <- llm_call(cfg$provider, msgs, max_tokens = 15L)
    recv <- first_valid_int(raw, valid_targets)
    if (is.na(recv)) recv <- label_match(raw, valid_targets, cfg)
    if (is.na(recv)) {
      warning("[llrem] roleplay_sender: could not parse '", raw,
              "'; substituting random target.", call. = FALSE)
      recv <- sample(valid_targets, 1L)
    }
    c(sender = sender_id, receiver = recv)
  }
}

#' Query factory: LLM predicts the next event
#'
#' Returns a `query_fn` closure that shows the LLM the event history and asks
#' it to predict both the sender and receiver of the next event. Tests the
#' model's implicit understanding of network dynamics rather than its own
#' behavioural tendencies.
#'
#' @param cfg A `rem_cfg` object.
#'
#' @return A function with signature
#'   `f(past_events, empirical_event, all_nodes) -> c(sender, receiver)`.
#' @examples
#' \dontrun{
#' cfg <- rem_cfg(
#'   provider    = llm_provider_openai(),
#'   context     = "Conflict events between states.",
#'   action_verb = "attacked",
#'   actor_noun  = "Country"
#' )
#' query_fn <- make_query_predict(cfg)
#' }
#' @export
make_query_predict <- function(cfg) {
  force(cfg)
  function(past_events, empirical_event, all_nodes) {
    node_str <- paste(
      paste0(node_label(all_nodes, cfg), " (", all_nodes, ")"), collapse = ", "
    )
    msgs <- make_messages(
      paste0(cfg$context,
             "\n\nObserve the sequence of past events and predict the next one. ",
             "Reply with EXACTLY two integers separated by a single space: ",
             "SENDER_ID RECEIVER_ID. No other text."),
      paste0("=== EVENT HISTORY ===\n", format_history(past_events, cfg),
             "\n\n=== PREDICTION TASK ===\n",
             "Actors: ", node_str, "\n",
             "Who acts next, and who do they target? Reply: SENDER_ID RECEIVER_ID")
    )
    raw  <- llm_call(cfg$provider, msgs, max_tokens = 25L)
    nums <- if (length(raw) == 0L || is.na(raw) || !nzchar(raw)) integer(0L) else suppressWarnings(
      as.integer(regmatches(raw, gregexpr("-?[0-9]+", raw))[[1L]])
    )
    nums     <- nums[nums %in% all_nodes]
    sender   <- if (length(nums) >= 1L) nums[1L] else label_match(raw, all_nodes, cfg)
    receiver <- if (length(nums) >= 2L) nums[2L] else NA_integer_
    valid_recv <- setdiff(all_nodes, sender)
    if (is.na(receiver) || identical(receiver, sender))
      receiver <- label_match(raw, valid_recv, cfg)
    if (is.na(sender) || is.na(receiver) || sender == receiver) {
      warning("[llrem] predict: could not parse '", raw,
              "'; substituting random event.", call. = FALSE)
      sender   <- sample(all_nodes, 1L)
      receiver <- sample(setdiff(all_nodes, sender), 1L)
    }
    c(sender = sender, receiver = receiver)
  }
}

#' Query factory: LLM roleplays as a randomly assigned node
#'
#' Returns a `query_fn` closure that assigns a uniformly random node identity
#' at each step and asks the LLM to choose a receiver. Provides a baseline
#' that isolates how much the assigned identity shapes LLM choices, independent
#' of the empirical sender sequence.
#'
#' @param cfg A `rem_cfg` object.
#'
#' @return A function with signature
#'   `f(past_events, empirical_event, all_nodes) -> c(sender, receiver)`.
#' @examples
#' \dontrun{
#' cfg <- rem_cfg(
#'   provider    = llm_provider_openai(),
#'   context     = "Conflict events between states.",
#'   action_verb = "attacked",
#'   actor_noun  = "Country"
#' )
#' query_fn <- make_query_roleplay_random(cfg)
#' }
#' @export
make_query_roleplay_random <- function(cfg) {
  force(cfg)
  inner <- make_query_roleplay_sender(cfg)
  function(past_events, empirical_event, all_nodes) {
    random_event        <- empirical_event
    random_event$sender <- sample(all_nodes, 1L)
    inner(past_events, random_event, all_nodes)
  }
}

test_that("run_llm_rem (roleplay_sender) returns correct structure", {
  res <- run_llm_rem(
    test_edgelist, test_formula,
    query_fn     = make_query_roleplay_sender(mock_cfg),
    n_nonevents  = NULL, base_riskset = test_riskset, seed = 1L
  )
  expect_named(res, c("sequence", "responses", "rem"))
  expect_s3_class(res$rem, "coxph")
})

test_that("run_llm_rem sequence has the correct number of rows", {
  res <- run_llm_rem(
    test_edgelist, test_formula,
    query_fn     = make_query_roleplay_sender(mock_cfg),
    n_nonevents  = NULL, base_riskset = test_riskset, seed = 1L
  )
  expect_equal(nrow(res$sequence), nrow(test_edgelist))
})

test_that("roleplay_sender preserves the empirical sender sequence", {
  res <- run_llm_rem(
    test_edgelist, test_formula,
    query_fn     = make_query_roleplay_sender(mock_cfg),
    n_nonevents  = NULL, base_riskset = test_riskset, seed = 1L
  )
  expect_identical(res$sequence$sender, test_edgelist$sender)
})

test_that("generated sequence contains no self-loops", {
  res <- run_llm_rem(
    test_edgelist, test_formula,
    query_fn     = make_query_roleplay_sender(mock_cfg),
    n_nonevents  = NULL, base_riskset = test_riskset, seed = 1L
  )
  expect_true(all(res$sequence$sender != res$sequence$receiver))
})

test_that("all generated receivers are valid node IDs", {
  res <- run_llm_rem(
    test_edgelist, test_formula,
    query_fn     = make_query_roleplay_sender(mock_cfg),
    n_nonevents  = NULL, base_riskset = test_riskset, seed = 1L
  )
  expect_true(all(res$sequence$receiver %in% test_nodes))
})

test_that("run_llm_rem (predict) produces no self-loops and valid IDs", {
  res <- run_llm_rem(
    test_edgelist, test_formula,
    query_fn     = make_query_predict(mock_cfg),
    n_nonevents  = NULL, base_riskset = test_riskset, seed = 2L
  )
  expect_true(all(res$sequence$sender   %in% test_nodes))
  expect_true(all(res$sequence$receiver %in% test_nodes))
  expect_true(all(res$sequence$sender   != res$sequence$receiver))
})

test_that("run_llm_rem (roleplay_random) senders deviate from empirical", {
  res <- run_llm_rem(
    test_edgelist, test_formula,
    query_fn     = make_query_roleplay_random(mock_cfg),
    n_nonevents  = NULL, base_riskset = test_riskset, seed = 3L
  )
  # Exclude the seeded first event; with 6 nodes at least one should differ
  n_diff <- sum(res$sequence$sender[-1L] != test_edgelist$sender[-1L])
  expect_gte(n_diff, 1L)
})

# fit_rem_sep -----------------------------------------------------------------

test_that("fit_rem_sep returns a rem_sep object with correct structure", {
  fit <- fit_rem_sep(
    test_edgelist,
    formula_init = survival::Surv(time1, occurred) ~
      RRecInitiation + NOTPSnd + survival::strata(time2),
    formula_cont = survival::Surv(time1, occurred) ~
      Repetition + RecRecencyLog + survival::strata(time2),
    n_nonevents  = NULL,
    base_riskset = test_riskset
  )
  expect_s3_class(fit, "rem_sep")
  expect_named(fit, c("init","cont","loglik","n_init","n_cont",
                       "init_fraction","method"), ignore.order = TRUE)
  expect_true(fit$n_init + fit$n_cont == nrow(test_edgelist) - 1L)
  expect_true(fit$init_fraction >= 0 && fit$init_fraction <= 1)
})

test_that("fit_rem_sep loglik is sum of sub-model logliks", {
  fit <- fit_rem_sep(
    test_edgelist,
    formula_init = survival::Surv(time1, occurred) ~
      RRecInitiation + survival::strata(time2),
    formula_cont = survival::Surv(time1, occurred) ~
      Repetition + survival::strata(time2),
    n_nonevents  = NULL,
    base_riskset = test_riskset
  )
  expect_equal(unname(fit$loglik["sep"]),
               unname(fit$loglik["init"]) + unname(fit$loglik["cont"]),
               tolerance = 1e-6)
})

test_that("fit_rem_sep initiation events are genuinely first contacts", {
  fit <- fit_rem_sep(
    test_edgelist,
    formula_init = survival::Surv(time1, occurred) ~
      RRecInitiation + survival::strata(time2),
    formula_cont = survival::Surv(time1, occurred) ~
      Repetition + survival::strata(time2),
    n_nonevents  = NULL,
    base_riskset = test_riskset
  )
  # n_init + n_cont == nrow(test_edgelist) - 1 (event 1 has no risk set)
  expect_equal(fit$n_init + fit$n_cont, nrow(test_edgelist) - 1L)
})

test_that("fit_rem_sep clogit method returns rem_clogit sub-models", {
  fit <- fit_rem_sep(
    test_edgelist,
    formula_init = survival::Surv(time1, occurred) ~
      RRecInitiation + survival::strata(time2),
    formula_cont = survival::Surv(time1, occurred) ~
      Repetition + survival::strata(time2),
    n_nonevents  = NULL,
    base_riskset = test_riskset,
    method       = "clogit"
  )
  expect_s3_class(fit, "rem_sep")
  if (!is.null(fit$init)) expect_s3_class(fit$init, "rem_clogit")
  if (!is.null(fit$cont)) expect_s3_class(fit$cont, "rem_clogit")
})

#' Create a prompt configuration object
#'
#' Bundles everything that affects prompt construction into a single reusable
#' object. One `rem_cfg` can be passed to multiple [make_query_roleplay_sender()],
#' [make_query_predict()], and [make_query_roleplay_random()] calls. Swap
#' `provider` to replicate a study with a different model.
#'
#' @param provider An `llm_provider` object created by one of the
#'   `llm_provider_*()` functions.
#' @param context One or two sentences describing the network domain in second
#'   person (used as the system message).
#' @param action_verb Verb describing a directed event, e.g. `"attacked"`,
#'   `"emailed"`, `"called"`.
#' @param actor_noun Capitalised singular noun for nodes, e.g. `"Country"`,
#'   `"Person"`, `"Actor"`.
#' @param node_labels Optional named character vector mapping integer node IDs
#'   (as strings) to readable labels, e.g. `c("1" = "USA", "2" = "Russia")`.
#'   `NULL` generates generic `"<actor_noun> <id>"` labels.
#' @param history_window Maximum number of past events included in each prompt.
#'   `NULL` includes the full history.
#'
#' @return A `rem_cfg` object.
#' @examples
#' \dontrun{
#' cfg <- rem_cfg(
#'   provider    = llm_provider_openai(),
#'   context     = "These are militarized conflict events between states.",
#'   action_verb = "attacked",
#'   actor_noun  = "Country",
#'   history_window = 20L
#' )
#' print(cfg)
#' }
#' @export
rem_cfg <- function(provider,
                    context,
                    action_verb,
                    actor_noun,
                    node_labels    = NULL,
                    history_window = 30L) {
  if (!inherits(provider, "llm_provider"))
    stop("`provider` must be an <llm_provider> object.", call. = FALSE)
  structure(
    list(provider = provider, context = context, action_verb = action_verb,
         actor_noun = actor_noun, node_labels = node_labels,
         history_window = history_window),
    class = "rem_cfg"
  )
}

#' @export
print.rem_cfg <- function(x, ...) {
  cat("<rem_cfg>\n")
  cat("  Provider :", x$provider$name, "/", x$provider$model, "\n")
  cat("  Action   : <actor>", x$action_verb, "<actor>\n")
  cat("  Noun     :", x$actor_noun, "\n")
  cat("  History  :",
      if (is.null(x$history_window)) "full" else x$history_window,
      "events\n")
  if (!is.null(x$node_labels))
    cat("  Labels   :", length(x$node_labels), "named nodes\n")
  invisible(x)
}

#' Forest plot comparing fitted REM coefficients
#'
#' Produces a coefficient plot with 95% confidence intervals for a named list
#' of `coxph` models, with one colour per model source.
#'
#' @param models Named list of `coxph` objects.
#' @param term_order Character vector giving the desired y-axis order (bottom
#'   to top). Defaults to reverse alphabetical over all terms present.
#' @param palette Named character vector of hex colours keyed to
#'   `names(models)`. A colour-blind-friendly default palette is applied when
#'   `NULL`.
#' @param title Plot title string.
#'
#' @return A `ggplot` object (invisibly). The plot is also printed as a
#'   side-effect.
#' @examples
#' \dontrun{
#' data(edgelist)
#' el <- prepare_edgelist(edgelist)
#' f  <- survival::Surv(time1, occurred) ~ Repetition + Recencylog +
#'          survival::strata(time2)
#' m1 <- fit_rem(el[1:150, ], f, n_nonevents = 30L, method = "cox")
#' m2 <- fit_rem(el[1:300, ], f, n_nonevents = 30L, method = "cox")
#' compare_rem_models(list("N=150" = m1, "N=300" = m2))
#' }
#' @export
compare_rem_models <- function(models,
                                term_order = NULL,
                                palette    = NULL,
                                title      = "REM Coefficient Comparison") {
  if (!is.list(models) || is.null(names(models)))
    stop("`models` must be a named list of coxph objects.", call. = FALSE)

  extract_one <- function(model, label) {
    s <- summary(model)$coefficients
    data.frame(term    = rownames(s),
               coef    = s[, "coef"],
               se      = s[, "se(coef)"],
               p       = s[, "Pr(>|z|)"],
               lower95 = s[, "coef"] - 1.96 * s[, "se(coef)"],
               upper95 = s[, "coef"] + 1.96 * s[, "se(coef)"],
               source  = label,
               row.names = NULL, stringsAsFactors = FALSE)
  }

  coef_df <- do.call(rbind, Map(extract_one, models, names(models)))

  if (is.null(term_order)) term_order <- rev(sort(unique(coef_df$term)))
  if (is.null(palette)) {
    defaults <- c("#2166AC", "#D6604D", "#4DAC26", "#A858A0",
                  "#F4A736", "#878787", "#01665E", "#8C510A")
    palette  <- stats::setNames(defaults[seq_along(models)], names(models))
  }

  coef_df$term   <- factor(coef_df$term,   levels = term_order)
  coef_df$source <- factor(coef_df$source, levels = names(models))

  p <- ggplot2::ggplot(
    coef_df,
    ggplot2::aes(x = coef, y = term, colour = source,
                 xmin = lower95, xmax = upper95)
  ) +
    ggplot2::geom_vline(xintercept = 0, linetype = "dashed", colour = "grey60") +
    ggplot2::geom_linerange(
      position = ggplot2::position_dodge(width = 0.6), linewidth = 0.8
    ) +
    ggplot2::geom_point(
      position = ggplot2::position_dodge(width = 0.6), size = 2.5
    ) +
    ggplot2::scale_colour_manual(values = palette) +
    ggplot2::labs(title = title,
                  x = "Coefficient (log hazard ratio)", y = NULL, colour = NULL) +
    ggplot2::theme_minimal(base_size = 13) +
    ggplot2::theme(legend.position = "bottom")

  print(p)
  invisible(p)
}

#' Extract sufficient-statistic means from a REM risk-set data frame
#'
#' Returns a data frame with one row per value of `occurred` (0/1), containing
#' the mean of each REM covariate.  Used as input to [make_behavior_queries()].
#'
#' @param rem_formula Survival formula whose RHS terms are the covariates.
#' @param data        Risk-set data frame (expanded counting-process format)
#'   with an `occurred` column.
#' @param drop_strata If `TRUE` (default), drop strata columns from the model
#'   matrix before computing means.
#'
#' @return A data frame with columns `occurred` plus one column per covariate.
#' @examples
#' data(edgelist)
#' el  <- prepare_edgelist(edgelist)
#' fit <- fit_rem(el[1:100, ],
#'   survival::Surv(time1, occurred) ~ Repetition + survival::strata(time2),
#'   n_nonevents = 30L, method = "clogit")
#' # make_suffstats operates on the expanded risk-set data frame; see fit_rem
#' @export
make_suffstats <- function(rem_formula, data, drop_strata = TRUE) {
  tt <- terms(rem_formula, specials = "strata", data = data)
  X  <- model.matrix(tt, data = data)

  if ("(Intercept)" %in% colnames(X))
    X <- X[, colnames(X) != "(Intercept)", drop = FALSE]

  if (drop_strata)
    X <- X[, !grepl("strata\\(", colnames(X)), drop = FALSE]

  as.data.frame(X)
}

#' Compute per-term mean sufficient statistics for realized events
#'
#' Aggregates a risk-set data frame by `occurred` and returns means of the REM
#' covariates, plus the median of `Recency` for realized events (if present).
#'
#' @param rem_formula Survival formula.
#' @param poss_edges  Risk-set data frame containing `occurred` and `Recency`.
#'
#' @return A data frame suitable for passing to [make_behavior_queries()].
#' @examples
#' \dontrun{
#' data(edgelist)
#' el  <- prepare_edgelist(edgelist)
#' f   <- survival::Surv(time1, occurred) ~ Repetition + SndInitiation +
#'          survival::strata(time2)
#' # poss_edges is the internal risk-set frame; obtain via fit_rem internals
#' }
#' @export
make_stat_means <- function(rem_formula, poss_edges) {
  ss <- make_suffstats(rem_formula, poss_edges)
  sm <- aggregate(
    . ~ occurred,
    data = cbind(occurred = poss_edges$occurred, ss),
    FUN  = mean
  )
  if ("Recency" %in% names(poss_edges)) {
    sm$RecencyMedian <- c(
      median(poss_edges$Recency[poss_edges$occurred == 0]),
      median(poss_edges$Recency[poss_edges$occurred == 1 &
                                  poss_edges$SndInitiation == 1])
    )
  }
  sm
}

#' Generate plain-language behavioral guidelines from empirical sufficient stats
#'
#' Translates the mean sufficient statistics for realized conflict events into
#' numbered natural-language instructions suitable for inclusion in an LLM
#' system prompt.  Intended to ground LLM predictions in the empirical base
#' rates of each relational pattern.
#'
#' @param stat_means  Data frame returned by [make_stat_means()]: one row per
#'   value of `occurred`, one column per REM term.
#' @param event_value The value of `occurred` that identifies realized events
#'   (default 1).
#' @param digits      Decimal digits for percentage rounding (default 0).
#'
#' @return A single character string: a numbered list of behavioral guidelines.
#' @examples
#' sm <- data.frame(
#'   occurred  = c(0, 1),
#'   Repetition = c(0.05, 0.45),
#'   IOviolation = c(0.12, 0.08)
#' )
#' cat(make_behavior_queries(sm))
#' @export
make_behavior_queries <- function(stat_means, event_value = 1, digits = 0) {
  event_row          <- stat_means[stat_means$occurred == event_value, ,
                                   drop = FALSE]
  event_row$occurred <- NULL

  pct <- function(x) paste0(round(100 * x, digits), "%")
  out <- character(0L)

  if ("Repetition" %in% names(event_row))
    out <- c(out, sprintf(
      "About %s of realized events from i to j had at least one prior i \u2192 j event. Prefer previously used i \u2192 j dyads over entirely new dyads.",
      pct(event_row$Repetition)
    ))

  if ("SndInitiation" %in% names(event_row))
    out <- c(out, sprintf(
      "About %s of realized events from i to j had at least one prior j \u2192 i event. Prefer targets that have previously attacked i.",
      pct(event_row$SndInitiation)
    ))

  if ("IOviolation" %in% names(event_row))
    out <- c(out, sprintf(
      "Only about %s of realized events are interaction-order violations -- events where a new actor targets the previous sender (AB-XA), the sender redirects to a new target (AB-XY),
      or a new actor targets the previous receiver (PSAB-XB). In this data, these participation shift patterns are strongly suppressed: after A attacks B, the sequence ten
ds NOT to jump to a new actor or a new target pairing. Reduce how often you introduce new participants or redirect aggression. For example: 69 attacked 10, but shortly after 38 attack
ed 69 (a new actor turning on the previous sender -- PSAB-XA). These participation shifts are less common than simple repetition.",
      pct(event_row$IOviolation)
    ))

  if ("Imitation" %in% names(event_row))
    out <- c(out, sprintf(
      "About %s of realized events involve imitation. Do not often choose a target simply because others recently attacked that target.",
      pct(event_row$Imitation)
    ))

  if ("NTriadic" %in% names(event_row))
    out <- c(out, sprintf(
      "About %s of realized events involve the triadic pattern. Use triadic closure only rarely.",
      pct(event_row$NTriadic)
    ))

  if ("RecRecencyLog" %in% names(event_row) && "RecencyMedian" %in% names(event_row))
    out <- c(out, sprintf(
      "Actors typically reciprocate an act within %s timepoints or less. Prefer i \u2192 j events that have had a recent j \u2192 i event.",
      event_row$RecencyMedian
    ))

  paste(seq_along(out), out, sep = ". ", collapse = "\n")
}

# Multi-agent REM-LLM simulation -----------------------------------------------

#' Compute the full hazard surface over all directed dyads
#'
#' Evaluates `exp(beta^T * x_ij)` for every directed dyad given the current
#' event history. Used internally by [run_multiagent_rem()] to derive sender
#' activation probabilities and the REM-hazard fallback for failed LLM parses.
#'
#' @param history     Data frame of past events (columns: time, sender,
#'   receiver, eveid).
#' @param base_riskset Precomputed base risk set (output of `get_base_riskset()`).
#' @param beta        Named numeric vector of REM coefficients (from
#'   `coef(rem_fit)`).
#' @param covariates Optional exogenous covariate list from [rem_covariates()].
#'
#' @return Data frame with columns `sender`, `receiver`, `hazard`.
#' @examples
#' data(edgelist)
#' el   <- prepare_edgelist(edgelist)
#' fit  <- fit_rem(el[1:100, ],
#'   survival::Surv(time1, occurred) ~ Repetition + survival::strata(time2),
#'   n_nonevents = 20L, method = "clogit")
#' nodes  <- sort(unique(c(el$sender, el$receiver)))
#' rs     <- expand.grid(sender = nodes, receiver = nodes)
#' rs     <- rs[rs$sender != rs$receiver, ]
#' rs$eid <- paste0(rs$sender, "_", rs$receiver)
#' beta <- c(Repetition = coef(fit)[["Repetition"]])
#' hz   <- compute_hazard_surface(el[1:5, ], rs, beta)
#' head(hz)
#' @export
compute_hazard_surface <- function(history, base_riskset, beta,
                                    covariates = NULL) {
  if (nrow(history) == 0L) {
    return(data.frame(sender   = base_riskset$sender,
                      receiver = base_riskset$receiver,
                      hazard   = 1,
                      stringsAsFactors = FALSE))
  }
  dummy_ev <- data.frame(
    time = max(history$time) + 1L, sender = -1L, receiver = -1L,
    eveid = "-1_-1", stringsAsFactors = FALSE
  )
  rs <- riskset_cov2(history, dummy_ev, n_nonevents = NULL,
                     base_riskset = base_riskset, covariates = covariates)
  for (i in seq_len(ncol(rs)))
    if (is.numeric(rs[[i]])) { rs[[i]][is.na(rs[[i]])] <- 0; rs[[i]][!is.finite(rs[[i]])] <- 0 }
  beta_ok <- beta[!is.na(beta)]
  cn <- intersect(names(beta_ok), colnames(rs))
  X  <- as.matrix(rs[, cn, drop = FALSE])
  lp <- if (length(cn) > 0L) drop(X %*% beta_ok[cn]) else rep(0, nrow(rs))
  data.frame(sender   = rs$sender,
             receiver = rs$receiver,
             hazard   = exp(lp),
             stringsAsFactors = FALSE)
}

# query_agent ------------------------------------------------------------------
# Ask one LLM agent to choose a receiver.  On parse failure, falls back to
# sampling from the sender's REM hazard row rather than uniform random.

query_agent <- function(sender_id, valid_targets, history, cfg, hazard_row) {
  target_str <- paste(
    paste0(node_label(valid_targets, cfg), " (", valid_targets, ")"),
    collapse = ", "
  )
  msgs <- make_messages(
    paste0(cfg$context,
           "\n\nYou are roleplaying as one actor in this network. ",
           "Given the event history, decide who you will target next. ",
           "Reply with ONLY the integer ID of your target -- no other text."),
    paste0("=== EVENT HISTORY ===\n", format_history(history, cfg),
           "\n\n=== YOUR TURN ===\n",
           "You are ", node_label(sender_id, cfg), " (ID: ", sender_id, ").\n",
           "Valid targets: ", target_str, "\n",
           "Which actor do you target? Reply with the integer ID only.")
  )
  raw  <- tryCatch(llm_call(cfg$provider, msgs, max_tokens = 15L),
                   error = function(e) NA_character_)
  recv <- first_valid_int(raw, valid_targets)
  if (is.na(recv)) recv <- label_match(raw, valid_targets, cfg)
  if (is.na(recv)) {
    h <- hazard_row
    h[is.na(h) | !is.finite(h)] <- 0
    if (sum(h) > 0) {
      recv <- sample(valid_targets, 1L, prob = h / sum(h))
    } else {
      recv <- sample(valid_targets, 1L)
    }
  }
  recv
}

#' Run a multi-agent REM-LLM conflict simulation
#'
#' At each tick, the REM hazard surface determines who acts (sender activation),
#' and individual LLM agents decide against whom (receiver selection). The two
#' components are complementary: the REM enforces endogenous structure; the LLM
#' contributes identity-conditioned content.
#'
#' @param edgelist        Prepared edgelist -- used to infer the node set and
#'   optionally as the seed event.
#' @param rem_fit         Fitted `coxph` object (from [fit_rem()]).
#' @param cfg             A single [rem_cfg()] shared across all agents.
#'   Sender identity is injected dynamically at each query.
#' @param n_ticks         Number of simulation ticks.
#' @param lambda          Mean events per tick (Poisson rate). Estimated from
#'   `edgelist` inter-event times if `NULL`.
#' @param base_riskset    Precomputed base risk set. Built from `edgelist` if
#'   `NULL`.
#' @param cold_start      If `TRUE`, begin with a uniform hazard surface rather
#'   than seeding from the first empirical event. Cold-start runs have a burn-in
#'   period before endogenous structure accumulates.
#' @param seed            RNG seed.
#' @param checkpoint_path File path for incremental saves every 10 ticks.
#'   `NULL` disables checkpointing. Existing checkpoint is resumed automatically.
#' @param covariates Optional exogenous covariate list from [rem_covariates()].
#'   Passed to [fit_rem()] when computing the hazard surface.
#'
#' @return A named list:
#'   \describe{
#'     \item{`sequence`}{Data frame of generated events (time, sender, receiver,
#'       eveid).}
#'     \item{`sender_probs`}{`n_ticks` x `N` matrix of activation probabilities
#'       \eqn{p_i(t)} at each tick.}
#'     \item{`n_events_per_tick`}{Integer vector of \eqn{K_t} values.}
#'   }
#' @examples
#' \dontrun{
#' data(edgelist)
#' el  <- prepare_edgelist(edgelist)
#' fit <- fit_rem(el[1:200, ],
#'   survival::Surv(time1, occurred) ~ Repetition + Recencylog +
#'     survival::strata(time2),
#'   n_nonevents = 50L, method = "cox")
#' cfg <- rem_cfg(
#'   provider    = llm_provider_openai(),
#'   context     = "Conflict events between states.",
#'   action_verb = "attacked",
#'   actor_noun  = "Country"
#' )
#' sim <- run_multiagent_rem(el[1:200, ], fit, cfg, n_ticks = 5L)
#' head(sim$sequence)
#' }
#' @export
run_multiagent_rem <- function(edgelist,
                                rem_fit,
                                cfg,
                                n_ticks,
                                lambda          = NULL,
                                base_riskset    = NULL,
                                cold_start      = FALSE,
                                seed            = 8214L,
                                checkpoint_path = NULL,
                                covariates      = NULL) {
  all_nodes    <- get_nodes(edgelist, NULL)
  base_riskset <- get_base_riskset(base_riskset, all_nodes)
  beta         <- coef(rem_fit)

  if (is.null(lambda)) {
    tr <- diff(range(edgelist$time))
    lambda <- if (tr > 0) nrow(edgelist) / tr else 1
  }

  set.seed(seed)

  # Initialise history and tick clock
  if (cold_start) {
    history   <- edgelist[0L, c("time", "sender", "receiver", "eveid")]
    tick_time <- 1L
  } else {
    history   <- edgelist[1L, c("time", "sender", "receiver", "eveid")]
    tick_time <- history$time[1L] + 1L
  }

  sequence_rows     <- list()
  sender_probs_list <- vector("list", n_ticks)
  n_events_per_tick <- integer(n_ticks)
  start_tick        <- 1L

  # Checkpoint recovery
  if (!is.null(checkpoint_path) && file.exists(checkpoint_path)) {
    env <- new.env(parent = emptyenv())
    load(checkpoint_path, envir = env)
    history           <- env$history
    sequence_rows     <- env$sequence_rows
    sender_probs_list <- env$sender_probs_list
    n_events_per_tick <- env$n_events_per_tick
    start_tick        <- env$next_tick
    tick_time         <- env$tick_time
    message("[llrem] Resuming multi-agent run from tick ", start_tick)
  }

  pb <- utils::txtProgressBar(min = start_tick, max = n_ticks, style = 3)

  for (tick in start_tick:n_ticks) {
    # Step 1 & 2: hazard surface -> sender activation probabilities
    hz  <- compute_hazard_surface(history, base_riskset, beta, covariates)
    H_i <- tapply(hz$hazard, hz$sender, sum)
    H_i <- H_i[as.character(all_nodes)]
    H_i[is.na(H_i)] <- 0
    total_H <- sum(H_i)
    p_i <- if (total_H > 0) H_i / total_H else rep(1 / length(all_nodes), length(all_nodes))
    sender_probs_list[[tick]] <- as.numeric(p_i)

    # Step 3: number of events this tick
    K_t <- stats::rpois(1L, lambda)
    if (K_t == 0L) {
      n_events_per_tick[tick] <- 0L
      tick_time <- tick_time + 1L
      utils::setTxtProgressBar(pb, tick)
      next
    }
    K_t <- min(K_t, length(all_nodes))
    n_events_per_tick[tick] <- K_t

    # Step 4: sample K_t senders without replacement
    sampled_senders <- sample(all_nodes, size = K_t, replace = FALSE, prob = p_i)

    # Step 5: query each agent (all see the same pre-tick history)
    tick_rows <- vector("list", K_t)
    for (s_idx in seq_len(K_t)) {
      sender_id     <- sampled_senders[s_idx]
      valid_targets <- setdiff(all_nodes, sender_id)
      hz_sub        <- hz[hz$sender == sender_id, ]
      hz_row        <- hz_sub$hazard[match(valid_targets, hz_sub$receiver)]
      hz_row[is.na(hz_row)] <- 0

      recv <- query_agent(sender_id, valid_targets, history, cfg,
                          hazard_row = hz_row)
      tick_rows[[s_idx]] <- data.frame(
        time = tick_time, sender = sender_id, receiver = recv,
        eveid = paste0(sender_id, "_", recv), stringsAsFactors = FALSE
      )
    }

    # Step 6: append all tick events (within-tick events invisible to co-senders)
    tick_df       <- do.call(rbind, tick_rows)
    sequence_rows <- c(sequence_rows, tick_rows)
    history       <- rbind(history, tick_df)
    tick_time     <- tick_time + 1L

    if (!is.null(checkpoint_path) && (tick %% 10L == 0L || tick == n_ticks)) {
      next_tick <- tick + 1L
      save(history, sequence_rows, sender_probs_list, n_events_per_tick,
           next_tick, tick_time, file = checkpoint_path)
    }

    utils::setTxtProgressBar(pb, tick)
  }
  close(pb)

  sequence <- do.call(rbind, sequence_rows)
  if (!cold_start && nrow(edgelist) >= 1L) {
    sequence <- rbind(edgelist[1L, c("time", "sender", "receiver", "eveid")],
                      sequence)
  }
  row.names(sequence) <- NULL

  sender_probs <- do.call(rbind, lapply(sender_probs_list, function(p) {
    if (is.null(p)) rep(NA_real_, length(all_nodes)) else p
  }))
  colnames(sender_probs) <- as.character(all_nodes)

  list(sequence          = sequence,
       sender_probs      = sender_probs,
       n_events_per_tick = n_events_per_tick)
}

#' Normalise a raw edgelist to the standard three-column format
#'
#' Accepts a matrix or data frame with columns in order: time, sender,
#' receiver. Returns a data frame with named integer columns and an `eveid`
#' dyad key (`"sender_receiver"`).
#'
#' @param raw Matrix or data frame with at least 3 columns.
#'
#' @return A data frame with columns `time` (integer), `sender` (integer),
#'   `receiver` (integer), and `eveid` (character).
#' @examples
#' el <- prepare_edgelist(data.frame(t = 1:5, s = c(1,2,3,1,2), r = c(2,3,1,3,1)))
#' head(el)
#' @export
prepare_edgelist <- function(raw) {
  df <- as.data.frame(raw)
  if (ncol(df) < 3L)
    stop("`raw` must have at least 3 columns: time, sender, receiver.",
         call. = FALSE)
  colnames(df)[1:3] <- c("time", "sender", "receiver")
  df$time     <- as.integer(df$time)
  df$sender   <- as.integer(df$sender)
  df$receiver <- as.integer(df$receiver)
  df$eveid    <- paste0(df$sender, "_", df$receiver)
  df
}

# Internal helpers ------------------------------------------------------------

get_nodes <- function(edgelist, n_hint = NULL) {
  nodes <- sort(unique(c(edgelist$sender, edgelist$receiver)))
  if (!is.null(n_hint) && n_hint > length(nodes)) nodes <- seq_len(n_hint)
  nodes
}

get_base_riskset <- function(base_riskset, all_nodes) {
  if (!is.null(base_riskset)) return(base_riskset)
  br     <- expand.grid(sender = all_nodes, receiver = all_nodes)
  br     <- br[br$sender != br$receiver, ]
  br$eid <- paste0(br$sender, "_", br$receiver)
  br
}

add_survival_cols <- function(pe) {
  pe$time1 <- 1L
  pe$time2 <- pe$time
  pe
}

# riskset_cov2 ----------------------------------------------------------------
# Compute REM covariates for one event given all past events.
#
# Covariates:
#   Repetition    dyad_occurrence / sender out-degree
#   Recency           time of last reverse-dyad event
#   RecRecencyLog     -log1p(Recency); 0 if j→i never fired
#   RecRecencyRank    1 / rank(j in i's reverse-dyad recency list); 0 if never fired
#   IOviolation   interaction-order violation (Simmelian sequence)
#   Imitation     (recInDg - dyad_occurrence) / per-sender max; cumulative in-degree of receiver from all other senders
#   OTP / ITP / OSP / ISP   raw two-path / shared-partner counts
#   NOTPSnd       OTP normalised by per-sender max (transitivity)
#   NITPSnd       ITP normalised by per-sender max (cyclic closure)
#   NOSPSnd       OSP normalised by per-sender max (outbound shared partners)
#   NISPSnd       ISP normalised by per-sender max (inbound shared partners)
#   NTriadic      normalised sum of all four (backward-compat composite)
#   OutDegSnd     normalised out-degree of sender:   senOutDg / (n-1)
#   OutDegRec     normalised out-degree of receiver: recOutDg / (n-1)
#   InDegSnd      normalised in-degree  of sender:   senInDg  / (n-1)
#   InDegRec      normalised in-degree  of receiver: recInDg  / (n-1)
#   TotDegSnd     normalised total degree of sender:   (senOutDg + senInDg) / (2*(n-1))
#   TotDegRec     normalised total degree of receiver: (recOutDg + recInDg) / (2*(n-1))
#   FrRecSnd      count(j→i prior events) / in-degree(i); cf. relevent FrRecSnd
#   PSABBA        p-shift AB→BA: was last event A→B and this is B→A? (turn receiving)
#   PSABBY        p-shift AB→BY: was last event A→B and this is B→Y? (turn receiving, new target)
#   PSABAY        p-shift AB→AY: was last event A→B and this is A→Y (Y≠B)? (turn continuing)
#   PSABXA        p-shift AB→XA: last event A→B, this is X→A (X≠B, X≠A)? (turn usurping, to A)
#   PSABXB        p-shift AB→XB: last event A→B, this is X→B (X≠A, X≠B)? (turn usurping, to B)
#   PSABXY        p-shift AB→XY: no shared sender/receiver with last event? (turn usurping, to new)
#   SndInitiation has this dyad occurred before? (binary)
#   RRecInitiation has the reverse dyad ever occurred? (reciprocity binary)
#   DyadRecencyLog  -log1p(time of last same-dyad event); 0 if never fired
#   WeightedRep      count of same-dyad events / per-sender max
#   WeightedRec      min(w_ij, w_ji) / max(w_ij, w_ji); Squartini et al. (2013)
#   occurred      1 for the focal event, 0 for non-events

riskset_cov2 <- function(past_events, current_event,
                          n_nonevents = NULL, base_riskset = NULL,
                          covariates = NULL) {
  riskset <- base_riskset
  ev      <- match(current_event$eveid, riskset$eid)

  if (!is.null(n_nonevents) && n_nonevents < nrow(riskset)) {
    keep    <- unique(c(ev, sample.int(nrow(riskset), n_nonevents)))
    riskset <- riskset[keep, ]
  }

  # Repetition
  outd                    <- table(past_events$sender)
  riskset$senOutDg        <- outd[match(riskset$sender, names(outd))]
  docc                    <- table(past_events$eveid)
  riskset$dyad_occurrence <- docc[match(riskset$eid, names(docc))]
  riskset$Repetition      <- riskset$dyad_occurrence / riskset$senOutDg

  # Recency
  last_time             <- tapply(past_events$time, past_events$eveid, max)
  recip_ids             <- paste0(riskset$receiver, "_", riskset$sender)
  riskset$Recency       <- last_time[match(recip_ids, names(last_time))]
  riskset$Recency[is.na(riskset$Recency)] <- 0
  riskset$RecRecencyLog <- -log1p(riskset$Recency)
  # RecRecencyRank: 1 / recency rank of j among senders to i (most recent = rank 1).
  # Analogous to relevent's RRecSnd: coef*(1/rank) contribution to i->j log-hazard.
  # Uses a deduplicated sender list ordered most-recent-first (same as relevent rrl).
  recip_senders <- split(past_events$sender,
                         past_events$receiver)[as.character(riskset$sender)]
  recip_times   <- split(past_events$time,
                         past_events$receiver)[as.character(riskset$sender)]
  riskset$RecRecencyRank <- mapply(function(snd_list, t_list, j) {
    if (is.null(snd_list) || length(snd_list) == 0L) return(0)
    # Deduplicate: keep only the most recent event per sender
    ord  <- order(t_list, decreasing = TRUE)
    uniq <- !duplicated(snd_list[ord])
    ranked_senders <- snd_list[ord][uniq]
    pos <- match(j, ranked_senders)
    if (is.na(pos)) 0 else 1 / pos
  }, recip_senders, recip_times, riskset$receiver,
  SIMPLIFY = TRUE, USE.NAMES = FALSE)

  # Interaction-order violation
  lastsender   <- utils::tail(past_events$sender,   1L)
  lastreceiver <- utils::tail(past_events$receiver, 1L)
  riskset$PSABAB <- riskset$sender   == lastsender   & riskset$receiver == lastreceiver
  riskset$PSABBA <- riskset$receiver == lastsender   & riskset$sender   == lastreceiver
  riskset$PSABAY <- riskset$sender   == lastsender   & !riskset$PSABAB
  riskset$PSABXA <- riskset$receiver == lastsender   & !riskset$PSABBA
  riskset$PSABXB <- riskset$receiver == lastreceiver & !riskset$PSABAB
  riskset$PSABBY <- riskset$sender   == lastreceiver & !riskset$PSABBA
  riskset$PSABXY <- (riskset$PSABXB + riskset$PSABXA + riskset$PSABAY +
                       riskset$PSABAB + riskset$PSABBA + riskset$PSABBY) == 0
  riskset$IOviolation <- riskset$PSABXA | riskset$PSABXY | riskset$PSABXB

  # Imitation
  indg                 <- table(past_events$receiver)
  riskset$recInDg      <- indg[match(riskset$receiver, names(indg))]
  riskset$PreImitation <- riskset$recInDg - riskset$dyad_occurrence
  riskset$PreImitation[is.na(riskset$PreImitation)] <- 0
  imimx                <- tapply(riskset$PreImitation, riskset$sender, max)
  imimx                <- imimx[match(riskset$sender, names(imimx))]
  riskset$Imitation    <- riskset$PreImitation / imimx

  # Triadic closure (OSP, ISP, ITP, OTP) via C++
  adj              <- build_adjacency(past_events$sender, past_events$receiver)
  counts           <- triadic_counts(adj, riskset$sender, riskset$receiver)
  riskset$OSP <- counts[, "OSP"]
  riskset$ISP <- counts[, "ISP"]
  riskset$ITP <- counts[, "ITP"]
  riskset$OTP <- counts[, "OTP"]

  # Per-sender-max normalisation (mirrors relevent's *Snd convention)
  norm_by_sender_max <- function(x, sender) {
    mx <- tapply(x, sender, max)
    mx <- mx[match(sender, names(mx))]
    ifelse(mx > 0, x / mx, 0)
  }
  riskset$NOTPSnd <- norm_by_sender_max(riskset$OTP, riskset$sender)
  riskset$NITPSnd <- norm_by_sender_max(riskset$ITP, riskset$sender)
  riskset$NOSPSnd <- norm_by_sender_max(riskset$OSP, riskset$sender)
  riskset$NISPSnd <- norm_by_sender_max(riskset$ISP, riskset$sender)

  riskset$triadic_closure <- riskset$OSP + riskset$ISP + riskset$ITP + riskset$OTP
  riskset$triadic_closure[is.na(riskset$triadic_closure)] <- 0L
  mx_tri           <- max(riskset$triadic_closure)
  riskset$NTriadic <- if (mx_tri > 0) riskset$triadic_closure / mx_tri else 0

  # Degree statistics scaled to [0,1] by (n-1).
  # Denominator is fixed at (n-1), the maximum possible degree given n nodes.
  # This differs from relevent's NODSnd/NODRec/NIDSnd/NIDRec, which divide by
  # the event-index `it` — a running count that changes each step.
  n_norm                  <- max(length(unique(c(base_riskset$sender,
                                                 base_riskset$receiver))) - 1L,
                                 1L)
  senInDg                 <- table(past_events$receiver)
  recOutDg                <- table(past_events$sender)
  riskset$OutDegSnd <- as.numeric(riskset$senOutDg) / n_norm
  riskset$OutDegSnd[is.na(riskset$OutDegSnd)] <- 0
  riskset$OutDegRec <- as.numeric(recOutDg[match(riskset$receiver, names(recOutDg))]) / n_norm
  riskset$OutDegRec[is.na(riskset$OutDegRec)] <- 0
  riskset$InDegSnd <- as.numeric(senInDg[match(riskset$sender, names(senInDg))]) / n_norm
  riskset$InDegSnd[is.na(riskset$InDegSnd)] <- 0
  riskset$InDegRec <- as.numeric(riskset$recInDg) / n_norm
  riskset$InDegRec[is.na(riskset$InDegRec)] <- 0

  # TotDegSnd / TotDegRec: (out-degree + in-degree) / (2*(n-1)).
  # Matches relevent's NTDegSnd/NTDegRec formula (same normalisation convention).
  riskset$TotDegSnd <- (riskset$OutDegSnd + riskset$InDegSnd) / 2
  riskset$TotDegRec <- (riskset$OutDegRec + riskset$InDegRec) / 2

  # Individual p-shift indicators (one-step-back patterns from last event).
  # IOviolation is the composite PSABXA | PSABXB | PSABXY (already present).
  # Expose the six individual indicators for users who want finer resolution.
  riskset$PSAB_BA <- as.numeric(riskset$PSABBA)
  riskset$PSAB_BY <- as.numeric(riskset$PSABBY)
  riskset$PSAB_AY <- as.numeric(riskset$PSABAY)
  riskset$PSAB_XA <- as.numeric(riskset$PSABXA)
  riskset$PSAB_XB <- as.numeric(riskset$PSABXB)
  riskset$PSAB_XY <- as.numeric(riskset$PSABXY)

  riskset$occurred        <- as.numeric(riskset$eid %in% current_event$eveid)
  riskset$time            <- current_event$time
  riskset$SndInitiation   <- as.numeric(riskset$dyad_occurrence > 0)
  riskset$RRecInitiation  <- as.numeric(!is.na(last_time[match(recip_ids, names(last_time))]))
  snd_last                <- last_time[match(riskset$eid, names(last_time))]
  snd_last[is.na(snd_last)] <- 0
  # DyadRecencyLog: -log1p(clock-time of last i→j event); 0 if never fired.
  # Not comparable to relevent's RSndSnd (a recency rank, opposite sign).
  riskset$DyadRecencyLog  <- -log1p(snd_last)

  # WeightedRep: count of same-dyad (i→j) events, normalised by per-sender max.
  # WeightedRec: weighted reciprocity per Squartini et al. (2013):
  #   min(w_ij, w_ji) / max(w_ij, w_ji); 0 when both are 0.
  rev_docc                <- docc[match(recip_ids, names(docc))]
  rev_docc[is.na(rev_docc)] <- 0L
  riskset$w_ji            <- as.integer(rev_docc)
  w_ij                    <- as.integer(riskset$dyad_occurrence)
  w_ij[is.na(w_ij)]      <- 0L
  riskset$WeightedRep     <- norm_by_sender_max(w_ij, riskset$sender)
  w_max                   <- pmax(w_ij, riskset$w_ji)
  w_min                   <- pmin(w_ij, riskset$w_ji)
  riskset$WeightedRec     <- ifelse(w_max > 0L, w_min / w_max, 0)

  # FrRecSnd: count(j→i past events) / in-degree(i); zero when in-degree is 0.
  # Matches relevent's FrRecSnd: fraction of i's past receipts that came from j.
  # Distinct from WeightedRep which normalises by per-sender max instead.
  sen_indeg               <- as.numeric(senInDg[match(riskset$sender, names(senInDg))])
  sen_indeg[is.na(sen_indeg)] <- 0
  riskset$FrRecSnd <- ifelse(sen_indeg > 0,
                              as.numeric(riskset$w_ji) / sen_indeg, 0)

  if (!is.null(covariates))
    riskset <- attach_covariates(riskset, covariates, current_event$time)

  riskset
}

# remstats_generate -----------------------------------------------------------

remstats_generate <- function(edgelist, n = NULL, n_nonevents = NULL,
                               cl = NULL, base_riskset = NULL,
                               covariates = NULL) {
  if (is.null(n))
    n <- length(unique(c(edgelist$sender, edgelist$receiver)))
  if (is.null(n_nonevents))
    n_nonevents <- n * (n - 1L)

  if (is.null(base_riskset)) {
    base_riskset     <- expand.grid(sender = seq_len(n), receiver = seq_len(n))
    base_riskset     <- base_riskset[base_riskset$sender != base_riskset$receiver, ]
  }
  base_riskset$eid <- paste0(base_riskset$sender, "_", base_riskset$receiver)
  edgelist$eveid   <- paste0(edgelist$sender, "_", edgelist$receiver)

  if (!is.null(cl)) {
    parallel::clusterExport(
      cl, c("n", "n_nonevents", "edgelist", "riskset_cov2", "base_riskset",
            "covariates", "attach_covariates", ".node_vals", ".find_window"),
      envir = environment()
    )
    pre <- parallel::parLapply(cl, seq(2L, nrow(edgelist)), \(y)
      riskset_cov2(edgelist[seq_len(y - 1L), ], edgelist[y, ],
                   n_nonevents, base_riskset, covariates))
  } else {
    pre <- vector("list", nrow(edgelist))
    for (y in seq(2L, nrow(edgelist)))
      pre[[y]] <- riskset_cov2(edgelist[seq_len(y - 1L), ], edgelist[y, ],
                                n_nonevents, base_riskset, covariates)
  }

  rs <- do.call(rbind.data.frame, pre)
  for (i in which(vapply(rs, is.logical, logical(1L))))
    rs[[i]] <- as.numeric(rs[[i]])

  # NA and non-finite fill applies only to endogenous columns; exogenous
  # covariate columns keep their NAs so the analyst can see missing data.
  exog_cols <- if (!is.null(covariates))
    unlist(lapply(covariates$covariates, function(cv) {
      if (inherits(cv, c("cov_node_static", "cov_node_temporal")) &&
          cv$role == "both")
        c(paste0(cv$name, "_snd"), paste0(cv$name, "_rec"))
      else
        cv$name
    }))
  else
    character(0L)

  endog_cols <- setdiff(names(rs), exog_cols)
  rs[endog_cols] <- lapply(rs[endog_cols], function(x) {
    if (is.numeric(x)) { x[is.na(x)] <- 0; x[!is.finite(x)] <- 0 }
    x
  })
  rs
}

test_that("prepare_edgelist returns a data frame with correct columns", {
  raw <- data.frame(t = 1:5, s = c(1,2,3,1,2), r = c(2,3,1,3,1))
  out <- prepare_edgelist(raw)
  expect_s3_class(out, "data.frame")
  expect_named(out, c("time", "sender", "receiver", "eveid"), ignore.order = FALSE)
})

test_that("prepare_edgelist coerces columns to integer", {
  raw <- data.frame(t = 1:3, s = c(1,2,3), r = c(2,3,1))
  out <- prepare_edgelist(raw)
  expect_type(out$time,     "integer")
  expect_type(out$sender,   "integer")
  expect_type(out$receiver, "integer")
})

test_that("prepare_edgelist forms eveid correctly", {
  raw <- data.frame(t = 1:2, s = c(1L, 3L), r = c(2L, 1L))
  out <- prepare_edgelist(raw)
  expect_equal(out$eveid, c("1_2", "3_1"))
})

test_that("prepare_edgelist accepts matrix input", {
  mat <- matrix(c(1,1,2, 2,2,3), nrow = 2, byrow = TRUE)
  out <- prepare_edgelist(mat)
  expect_s3_class(out, "data.frame")
  expect_true("eveid" %in% names(out))
})

test_that("prepare_edgelist errors on fewer than 3 columns", {
  expect_snapshot(prepare_edgelist(data.frame(a = 1:3, b = 1:3)), error = TRUE)
})

test_that("fit_rem returns a coxph object", {
  m <- fit_rem(test_edgelist, test_formula,
               n_nonevents = NULL, base_riskset = test_riskset)
  expect_s3_class(m, "coxph")
})

test_that("fit_rem produces expected coefficient names", {
  m <- fit_rem(test_edgelist, test_formula,
               n_nonevents = NULL, base_riskset = test_riskset)
  expected <- c("Imitation", "Repetition", "SndInitiation",
                "RecRecencyLog", "IOviolation", "NTriadic")
  # coef() includes strata terms; check that the core covariates are present
  expect_true(all(expected %in% names(stats::coef(m))))
})

test_that("fit_rem coefficients are all finite", {
  m <- fit_rem(test_edgelist, test_formula,
               n_nonevents = NULL, base_riskset = test_riskset)
  expect_true(all(is.finite(stats::coef(m))))
})

test_that("fit_rem errors on a 0-row edgelist", {
  expect_snapshot(
    fit_rem(test_edgelist[0L, ], test_formula, base_riskset = test_riskset),
    error = TRUE
  )
})

# Triadic closure (C++ implementation) ----------------------------------------

test_that("triadic_counts returns correct OSP/ISP/ITP/OTP for a known graph", {
  # Triangle: 1->2, 2->3, 1->3
  # Node 1 out-neighbours: {2, 3}
  # Node 2 out-neighbours: {3}
  # Node 3 out-neighbours: {}
  # Node 1 in-neighbours:  {}
  # Node 2 in-neighbours:  {1}
  # Node 3 in-neighbours:  {1, 2}
  snd <- c(1L, 2L, 1L)
  rec <- c(2L, 3L, 3L)
  adj <- llrem:::build_adjacency(snd, rec)

  # For dyad 1->2:
  #   OSP: intersect(out[1], out[2]) = intersect({2,3},{3}) = {3} -> 1
  #   ISP: intersect(in[1],  in[2])  = intersect({},  {1}) = {}  -> 0
  #   ITP: intersect(in[1],  out[2]) = intersect({},  {3}) = {}  -> 0
  #   OTP: intersect(out[1], in[2])  = intersect({2,3},{1}) = {} -> 0
  counts <- llrem:::triadic_counts(adj, 1L, 2L)
  expect_equal(as.integer(counts[, "OSP"]), 1L)
  expect_equal(as.integer(counts[, "ISP"]), 0L)
  expect_equal(as.integer(counts[, "ITP"]), 0L)
  expect_equal(as.integer(counts[, "OTP"]), 0L)

  # For dyad 2->3:
  #   OSP: intersect(out[2], out[3]) = intersect({3},{}) = {} -> 0
  #   ISP: intersect(in[2],  in[3])  = intersect({1},{1,2}) = {1} -> 1
  #   ITP: intersect(in[2],  out[3]) = intersect({1},{})  = {} -> 0
  #   OTP: intersect(out[2], in[3])  = intersect({3},{1,2}) = {} -> 0
  counts2 <- llrem:::triadic_counts(adj, 2L, 3L)
  expect_equal(as.integer(counts2[, "ISP"]), 1L)
  expect_equal(as.integer(counts2[, "OSP"]), 0L)
})

test_that("NTriadic from C++ matches igraph reference on small graph", {
  skip_if_not_installed("igraph")
  # Build a small edgelist with enough events for triadic stats
  el <- prepare_edgelist(data.frame(
    t = 1:6,
    s = c(1L, 2L, 3L, 1L, 2L, 4L),
    r = c(2L, 3L, 1L, 3L, 1L, 1L)
  ))
  rs <- llrem:::get_base_riskset(NULL, llrem:::get_nodes(el))

  # C++ path (current implementation)
  cpp_rs <- llrem:::riskset_cov2(el[1:5, ], el[6, ], n_nonevents = NULL,
                                  base_riskset = rs)

  # Reference: compute triadic closure with igraph directly
  past <- el[1:5, ]
  g    <- igraph::graph_from_data_frame(past[, c("sender", "receiver")],
                                         directed = TRUE)
  ref_osp <- integer(nrow(rs))
  ref_isp <- integer(nrow(rs))
  ref_itp <- integer(nrow(rs))
  ref_otp <- integer(nrow(rs))
  in_g <- which(rs$sender %in% igraph::V(g) & rs$receiver %in% igraph::V(g))
  lookup    <- unique(c(rs$sender[in_g], rs$receiver[in_g]))
  outneighs <- lapply(lookup, \(a) igraph::neighbors(g, a, mode = "out"))
  inneighs  <- lapply(lookup, \(a) igraph::neighbors(g, a, mode = "in"))
  names(outneighs) <- names(inneighs) <- as.character(lookup)
  for (i in in_g) {
    a <- as.character(rs$sender[i]);  b <- as.character(rs$receiver[i])
    ref_osp[i] <- length(intersect(outneighs[[a]], outneighs[[b]]))
    ref_isp[i] <- length(intersect(inneighs[[a]],  inneighs[[b]]))
    ref_itp[i] <- length(intersect(inneighs[[a]],  outneighs[[b]]))
    ref_otp[i] <- length(intersect(outneighs[[a]], inneighs[[b]]))
  }

  expect_equal(cpp_rs$OSP, ref_osp)
  expect_equal(cpp_rs$ISP, ref_isp)
  expect_equal(cpp_rs$ITP, ref_itp)
  expect_equal(cpp_rs$OTP, ref_otp)
})

# Disaggregated normalised triadic terms --------------------------------------

test_that("NOTPSnd / NITPSnd / NOSPSnd / NISPSnd are present and in [0, 1]", {
  el <- prepare_edgelist(data.frame(
    t = 1:8,
    s = c(1L, 2L, 3L, 1L, 2L, 3L, 4L, 1L),
    r = c(2L, 3L, 1L, 3L, 4L, 2L, 1L, 4L)
  ))
  rs  <- llrem:::get_base_riskset(NULL, llrem:::get_nodes(el))
  out <- llrem:::riskset_cov2(el[1:7, ], el[8, ], n_nonevents = NULL,
                               base_riskset = rs)
  for (col in c("NOTPSnd", "NITPSnd", "NOSPSnd", "NISPSnd")) {
    expect_true(col %in% names(out), label = paste(col, "present"))
    expect_true(all(out[[col]] >= 0 & out[[col]] <= 1),
                label = paste(col, "in [0,1]"))
  }
})

test_that("NOTPSnd is zero wherever OTP is zero", {
  el <- prepare_edgelist(data.frame(
    t = 1:8,
    s = c(1L, 2L, 3L, 1L, 2L, 3L, 4L, 1L),
    r = c(2L, 3L, 1L, 3L, 4L, 2L, 1L, 4L)
  ))
  rs  <- llrem:::get_base_riskset(NULL, llrem:::get_nodes(el))
  out <- llrem:::riskset_cov2(el[1:7, ], el[8, ], n_nonevents = NULL,
                               base_riskset = rs)
  expect_true(all(out$NOTPSnd[out$OTP == 0] == 0))
  expect_true(all(out$NOSPSnd[out$OSP == 0] == 0))
})

test_that("per-sender max normalisation: max NOTPSnd per sender is 0 or 1", {
  el <- prepare_edgelist(data.frame(
    t = 1:8,
    s = c(1L, 2L, 3L, 1L, 2L, 3L, 4L, 1L),
    r = c(2L, 3L, 1L, 3L, 4L, 2L, 1L, 4L)
  ))
  rs  <- llrem:::get_base_riskset(NULL, llrem:::get_nodes(el))
  out <- llrem:::riskset_cov2(el[1:7, ], el[8, ], n_nonevents = NULL,
                               base_riskset = rs)
  sender_maxes <- tapply(out$NOTPSnd, out$sender, max)
  expect_true(all(sender_maxes == 0 | sender_maxes == 1))
})

#' Militarized Interstate Dispute conflict edgelist
#'
#' A directed event sequence of militarized conflict acts between sovereign
#' states, drawn from the Militarized Interstate Dispute (MID) dataset.
#' Each row records one country using military force against another.
#'
#' @format A data frame with 1,708 rows and 3 variables:
#' \describe{
#'   \item{ev}{Integer event index (1--1708).}
#'   \item{id_snd}{Integer sender country ID.}
#'   \item{id_rec}{Integer receiver country ID.}
#' }
#'
#' @source Militarized Interstate Dispute (MID) dataset.
#' @usage data(edgelist)
"edgelist"

#' Number of nodes in the MID edgelist
#'
#' Total number of unique country nodes in the bundled MID conflict edgelist
#' (95 countries).
#'
#' @format An integer scalar.
#' @source Derived from the MID dataset bundled in this package.
#' @usage data(n_nodes)
"n_nodes"

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

build_adjacency <- function(senders, receivers) {
    .Call(`_llrem_build_adjacency`, senders, receivers)
}

triadic_counts <- function(adj, rs_senders, rs_receivers) {
    .Call(`_llrem_triadic_counts`, adj, rs_senders, rs_receivers)
}

#' Create a mock LLM provider for testing
#'
#' Returns an `llm_provider` that calls a local `strategy_fn` instead of
#' making any network requests. Useful for testing pipeline logic and custom
#' `query_fn` closures without incurring API costs.
#'
#' Three ready-made strategies are provided:
#' - [mock_strategy_min_id()]: always picks the smallest valid target ID.
#' - [mock_strategy_max_id()]: always picks the largest valid target ID.
#' - [mock_strategy_highest_indegree()]: picks the valid target with the
#'   highest cumulative in-degree in the event history.
#'
#' @param strategy_fn Function with signature
#'   `f(messages, max_tokens, temperature) -> character`. The return value
#'   must be parseable by the corresponding query factory (a stringified
#'   integer for roleplay scenarios; two space-separated integers for the
#'   predict scenario).
#' @param name Display name for `print()` output.
#'
#' @return An `llm_provider` object.
#' @examples
#' p <- llm_provider_mock(mock_strategy_min_id)
#' print(p)
#' @export
llm_provider_mock <- function(strategy_fn, name = "Mock") {
  force(strategy_fn)
  new_provider(name, "mock",
    \(messages, max_tokens = 15L, temperature = 0.7)
      strategy_fn(messages, max_tokens, temperature)
  )
}

#' Mock strategy: always pick the smallest valid target
#'
#' @param messages List of message objects passed by the provider interface.
#' @param max_tokens Maximum tokens (ignored by mock; present for interface
#'   compatibility).
#' @param temperature Sampling temperature (ignored by mock; present for
#'   interface compatibility).
#' @return A strategy function for use with [llm_provider_mock()].
#' @examples
#' msgs <- list(list(role = "user",
#'   content = "Valid targets: (2) (3) (4)"))
#' mock_strategy_min_id(msgs, 15L, 0.7)
#' @export
mock_strategy_min_id <- function(messages, max_tokens, temperature) {
  ids <- parse_valid_ids(messages)
  if (length(ids) == 0L) return("1")
  if (is_predict_msg(messages)) {
    s <- min(ids); r <- min(ids[ids != s])
    return(paste(s, r))
  }
  as.character(min(ids))
}

#' Mock strategy: always pick the largest valid target
#'
#' @param messages List of message objects passed by the provider interface.
#' @param max_tokens Maximum tokens (ignored by mock; present for interface
#'   compatibility).
#' @param temperature Sampling temperature (ignored by mock; present for
#'   interface compatibility).
#' @return A strategy function for use with [llm_provider_mock()].
#' @examples
#' msgs <- list(list(role = "user",
#'   content = "Valid targets: (2) (3) (4)"))
#' mock_strategy_max_id(msgs, 15L, 0.7)
#' @export
mock_strategy_max_id <- function(messages, max_tokens, temperature) {
  ids <- parse_valid_ids(messages)
  if (length(ids) == 0L) return("1")
  if (is_predict_msg(messages)) {
    s <- max(ids); r <- max(ids[ids != s])
    return(paste(s, r))
  }
  as.character(max(ids))
}

#' Mock strategy: pick the valid target with the highest in-degree
#'
#' Parses the event history embedded in the prompt to tally cumulative
#' in-degrees, then returns the valid target node with the highest count.
#' Falls back to the smallest valid ID when no history is available yet.
#'
#' @param messages List of message objects passed by the provider interface.
#' @param max_tokens Maximum tokens (ignored by mock; present for interface
#'   compatibility).
#' @param temperature Sampling temperature (ignored by mock; present for
#'   interface compatibility).
#' @return A strategy function for use with [llm_provider_mock()].
#' @examples
#' msgs <- list(list(role = "user", content = paste(
#'   "Event 1: Actor 1 (1) attacked Actor 2 (2).",
#'   "Event 2: Actor 1 (1) attacked Actor 2 (2).",
#'   "Valid targets: (2) (3) (4)", sep = "\n"
#' )))
#' mock_strategy_highest_indegree(msgs, 15L, 0.7)
#' @export
mock_strategy_highest_indegree <- function(messages, max_tokens, temperature) {
  ids   <- parse_valid_ids(messages)
  if (length(ids) == 0L) return("1")
  indeg <- parse_indegree(messages)

  pick_best <- function(candidates) {
    if (length(indeg) > 0L) {
      ranked <- as.integer(names(indeg))
      best   <- ranked[ranked %in% candidates]
      if (length(best) > 0L) return(best[1L])
    }
    min(candidates)
  }

  if (is_predict_msg(messages)) {
    sender   <- pick_best(ids)
    receiver <- pick_best(ids[ids != sender])
    paste(sender, receiver)
  } else {
    as.character(pick_best(ids))
  }
}

# Message parsing helpers (internal) ------------------------------------------

user_content <- function(messages) {
  for (m in messages) if (identical(m$role, "user")) return(m$content)
  ""
}

parens_ints <- function(text) {
  m <- regmatches(text, gregexpr("(?<=\\()\\d+(?=\\))", text, perl = TRUE))[[1L]]
  suppressWarnings(as.integer(m))
}

is_predict_msg <- function(messages) {
  grepl("SENDER_ID RECEIVER_ID", user_content(messages), fixed = TRUE)
}

parse_valid_ids <- function(messages) {
  lines <- strsplit(user_content(messages), "\n", fixed = TRUE)[[1L]]
  tline <- grep("Valid targets:|Actors:", lines, value = TRUE)
  if (length(tline) == 0L) return(integer(0L))
  parens_ints(tline[1L])
}

parse_indegree <- function(messages) {
  lines   <- strsplit(user_content(messages), "\n", fixed = TRUE)[[1L]]
  evlines <- grep("Event \\d+:", lines, value = TRUE)
  if (length(evlines) == 0L) return(integer(0L))
  recvs <- vapply(evlines, function(ln) {
    ids <- parens_ints(ln)
    if (length(ids) >= 2L) ids[2L] else NA_integer_
  }, integer(1L), USE.NAMES = FALSE)
  sort(table(recvs[!is.na(recvs)]), decreasing = TRUE)
}

# prepare_edgelist errors on fewer than 3 columns

    Code
      prepare_edgelist(data.frame(a = 1:3, b = 1:3))
    Condition
      Error:
      ! `raw` must have at least 3 columns: time, sender, receiver.

# fit_rem errors on a 0-row edgelist

    Code
      fit_rem(test_edgelist[0L, ], test_formula, base_riskset = test_riskset)
    Condition
      Error in `$<-.data.frame`:
      ! replacement has 1 row, data has 0

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rem.R
\name{fit_rem_sep}
\alias{fit_rem_sep}
\title{Fit a Separable Relational Event Model}
\usage{
fit_rem_sep(
  edgelist,
  formula_init,
  formula_cont,
  n = NULL,
  n_nonevents = 200L,
  base_riskset = NULL,
  cl = NULL,
  method = c("cox", "clogit", "temporal"),
  covariates = NULL,
  min_events = 2L
)
}
\arguments{
\item{edgelist}{Data frame with columns \code{time}, \code{sender}, \code{receiver},
\code{eveid}. Use \code{\link[=prepare_edgelist]{prepare_edgelist()}} to create this from a raw matrix.}

\item{formula_init}{Survival formula for the initiation sub-model
(events where \code{SndInitiation == 0}).  Typical terms: \code{RRecInitiation},
\code{NOTPSnd}, \code{NITPSnd}.  Must include \code{survival::strata(time2)}.}

\item{formula_cont}{Survival formula for the continuation sub-model
(events where \code{SndInitiation == 1}).  Typical terms: \code{Repetition},
\code{RecRecencyLog}.  Must include \code{survival::strata(time2)}.}

\item{n}{Number of nodes. Inferred from \code{edgelist} if \code{NULL}.}

\item{n_nonevents}{Sampled non-event risk-set size per event. \code{NULL} uses
the full risk set. Passed to \code{\link[=fit_rem]{fit_rem()}} internally.}

\item{base_riskset}{Precomputed base risk-set data frame.}

\item{cl}{Optional PSOCK cluster for parallelised risk-set computation.}

\item{method}{Estimation method: \code{"cox"} (default), \code{"clogit"}, or
\code{"temporal"}. Applied independently to each sub-model.}

\item{covariates}{Optional exogenous covariate list from \code{\link[=rem_covariates]{rem_covariates()}}.}

\item{min_events}{Minimum number of focal events required in each partition
to attempt fitting. Partitions below this threshold return \code{NULL} for
that sub-model with a warning. Default 2.}
}
\value{
A \code{rem_sep} object (a named list) with elements:
\describe{
\item{\code{init}}{Fitted initiation sub-model (same class as \code{\link[=fit_rem]{fit_rem()}}
output for the chosen \code{method}), or \code{NULL} if too few events.}
\item{\code{cont}}{Fitted continuation sub-model, or \code{NULL}.}
\item{\code{loglik}}{Named numeric vector: \code{init}, \code{cont}, \code{sep}
(= init + cont), and \code{null} (= init_null + cont_null).}
\item{\code{n_init}}{Number of initiation events in the sequence.}
\item{\code{n_cont}}{Number of continuation events.}
\item{\code{init_fraction}}{Proportion of events that are initiations.}
\item{\code{method}}{The estimation method used.}
}
}
\description{
Partitions the risk set by prior contact status (\code{SndInitiation}) and fits
two independent sub-models: one for \strong{initiation} events (sender has never
previously contacted receiver; \code{SndInitiation == 0}) and one for
\strong{continuation} events (sender has contacted receiver before;
\code{SndInitiation == 1}).  The approach is analogous to separable TERGMs
(Krivitsky and Handcock, 2014): because the two subsets are disjoint, their
parameters are conditionally independent and can be estimated separately via
standard Cox partial likelihood on each partition.
}
\details{
The combined log-likelihood is \code{loglik_init + loglik_cont}, which is directly
comparable to a pooled five-parameter model with the same degrees of freedom
(see paper for details).
}
\examples{
data(edgelist)
el <- prepare_edgelist(edgelist)
fit <- fit_rem_sep(
  el[1:200, ],
  formula_init = survival::Surv(time1, occurred) ~
    RRecInitiation + survival::strata(time2),
  formula_cont = survival::Surv(time1, occurred) ~
    Repetition + RecRecencyLog + survival::strata(time2),
  n_nonevents = 20L, method = "clogit"
)
print(fit)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/covariates.R
\name{cov_dyad_temporal}
\alias{cov_dyad_temporal}
\title{Create a temporal dyad-level covariate}
\usage{
cov_dyad_temporal(name, data, warn_missing = TRUE)
}
\arguments{
\item{name}{Character scalar.}

\item{data}{Data frame with columns \code{sender}, \code{receiver}, \code{time_start},
and \code{value}. Rows need not be sorted.}

\item{warn_missing}{Logical.}
}
\value{
An object of class \code{c("cov_dyad_temporal", "rem_cov")}.
}
\description{
Like \code{\link[=cov_dyad_static]{cov_dyad_static()}} but values change at known time breakpoints.
}
\examples{
treaty <- data.frame(
  sender = c(1L, 1L), receiver = c(2L, 2L),
  time_start = c(0, 15), value = c(0, 1)
)
cov <- cov_dyad_temporal("Treaty", treaty)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{rem_cfg}
\alias{rem_cfg}
\title{Create a prompt configuration object}
\usage{
rem_cfg(
  provider,
  context,
  action_verb,
  actor_noun,
  node_labels = NULL,
  history_window = 30L
)
}
\arguments{
\item{provider}{An \code{llm_provider} object created by one of the
\verb{llm_provider_*()} functions.}

\item{context}{One or two sentences describing the network domain in second
person (used as the system message).}

\item{action_verb}{Verb describing a directed event, e.g. \code{"attacked"},
\code{"emailed"}, \code{"called"}.}

\item{actor_noun}{Capitalised singular noun for nodes, e.g. \code{"Country"},
\code{"Person"}, \code{"Actor"}.}

\item{node_labels}{Optional named character vector mapping integer node IDs
(as strings) to readable labels, e.g. \code{c("1" = "USA", "2" = "Russia")}.
\code{NULL} generates generic \code{"<actor_noun> <id>"} labels.}

\item{history_window}{Maximum number of past events included in each prompt.
\code{NULL} includes the full history.}
}
\value{
A \code{rem_cfg} object.
}
\description{
Bundles everything that affects prompt construction into a single reusable
object. One \code{rem_cfg} can be passed to multiple \code{\link[=make_query_roleplay_sender]{make_query_roleplay_sender()}},
\code{\link[=make_query_predict]{make_query_predict()}}, and \code{\link[=make_query_roleplay_random]{make_query_roleplay_random()}} calls. Swap
\code{provider} to replicate a study with a different model.
}
\examples{
\dontrun{
cfg <- rem_cfg(
  provider    = llm_provider_openai(),
  context     = "These are militarized conflict events between states.",
  action_verb = "attacked",
  actor_noun  = "Country",
  history_window = 20L
)
print(cfg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_correction_query.R
\name{make_behavior_queries}
\alias{make_behavior_queries}
\title{Generate plain-language behavioral guidelines from empirical sufficient stats}
\usage{
make_behavior_queries(stat_means, event_value = 1, digits = 0)
}
\arguments{
\item{stat_means}{Data frame returned by \code{\link[=make_stat_means]{make_stat_means()}}: one row per
value of \code{occurred}, one column per REM term.}

\item{event_value}{The value of \code{occurred} that identifies realized events
(default 1).}

\item{digits}{Decimal digits for percentage rounding (default 0).}
}
\value{
A single character string: a numbered list of behavioral guidelines.
}
\description{
Translates the mean sufficient statistics for realized conflict events into
numbered natural-language instructions suitable for inclusion in an LLM
system prompt.  Intended to ground LLM predictions in the empirical base
rates of each relational pattern.
}
\examples{
sm <- data.frame(
  occurred  = c(0, 1),
  Repetition = c(0.05, 0.45),
  IOviolation = c(0.12, 0.08)
)
cat(make_behavior_queries(sm))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{edgelist}
\alias{edgelist}
\title{Militarized Interstate Dispute conflict edgelist}
\format{
A data frame with 1,708 rows and 3 variables:
\describe{
\item{ev}{Integer event index (1--1708).}
\item{id_snd}{Integer sender country ID.}
\item{id_rec}{Integer receiver country ID.}
}
}
\source{
Militarized Interstate Dispute (MID) dataset.
}
\usage{
data(edgelist)
}
\description{
A directed event sequence of militarized conflict acts between sovereign
states, drawn from the Militarized Interstate Dispute (MID) dataset.
Each row records one country using military force against another.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_correction_query.R
\name{make_stat_means}
\alias{make_stat_means}
\title{Compute per-term mean sufficient statistics for realized events}
\usage{
make_stat_means(rem_formula, poss_edges)
}
\arguments{
\item{rem_formula}{Survival formula.}

\item{poss_edges}{Risk-set data frame containing \code{occurred} and \code{Recency}.}
}
\value{
A data frame suitable for passing to \code{\link[=make_behavior_queries]{make_behavior_queries()}}.
}
\description{
Aggregates a risk-set data frame by \code{occurred} and returns means of the REM
covariates, plus the median of \code{Recency} for realized events (if present).
}
\examples{
\dontrun{
data(edgelist)
el  <- prepare_edgelist(edgelist)
f   <- survival::Surv(time1, occurred) ~ Repetition + SndInitiation +
         survival::strata(time2)
# poss_edges is the internal risk-set frame; obtain via fit_rem internals
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{n_nodes}
\alias{n_nodes}
\title{Number of nodes in the MID edgelist}
\format{
An integer scalar.
}
\source{
Derived from the MID dataset bundled in this package.
}
\usage{
data(n_nodes)
}
\description{
Total number of unique country nodes in the bundled MID conflict edgelist
(95 countries).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/covariates.R
\name{cov_node_temporal}
\alias{cov_node_temporal}
\title{Create a temporal node-level covariate}
\usage{
cov_node_temporal(
  name,
  data,
  role = c("both", "sender", "receiver"),
  warn_missing = TRUE
)
}
\arguments{
\item{name}{Character scalar.}

\item{data}{Data frame with columns \code{node} (integer-coercible),
\code{time_start} (numeric), and \code{value} (numeric). Rows need not be
sorted; sorting is done internally.}

\item{role}{One of \code{"both"} (default), \code{"sender"}, or \code{"receiver"}.}

\item{warn_missing}{Logical.}
}
\value{
An object of class \code{c("cov_node_temporal", "rem_cov")}.
}
\description{
Like \code{\link[=cov_node_static]{cov_node_static()}} but values change at known time breakpoints.
The active value at event time \code{t} is the one with the largest
\code{time_start <= t} (step function).
}
\examples{
sanctions <- data.frame(
  node = c(1L, 1L, 2L), time_start = c(0, 10, 5), value = c(0, 1, 1)
)
cov <- cov_node_temporal("Sanctioned", sanctions, role = "receiver")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock.R
\name{mock_strategy_highest_indegree}
\alias{mock_strategy_highest_indegree}
\title{Mock strategy: pick the valid target with the highest in-degree}
\usage{
mock_strategy_highest_indegree(messages, max_tokens, temperature)
}
\arguments{
\item{messages}{List of message objects passed by the provider interface.}

\item{max_tokens}{Maximum tokens (ignored by mock; present for interface
compatibility).}

\item{temperature}{Sampling temperature (ignored by mock; present for
interface compatibility).}
}
\value{
A strategy function for use with \code{\link[=llm_provider_mock]{llm_provider_mock()}}.
}
\description{
Parses the event history embedded in the prompt to tally cumulative
in-degrees, then returns the valid target node with the highest count.
Falls back to the smallest valid ID when no history is available yet.
}
\examples{
msgs <- list(list(role = "user", content = paste(
  "Event 1: Actor 1 (1) attacked Actor 2 (2).",
  "Event 2: Actor 1 (1) attacked Actor 2 (2).",
  "Valid targets: (2) (3) (4)", sep = "\n"
)))
mock_strategy_highest_indegree(msgs, 15L, 0.7)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rem.R
\name{compute_seq_stat_means}
\alias{compute_seq_stat_means}
\title{Compute sufficient-statistic means for a sequence without fitting a REM}
\usage{
compute_seq_stat_means(
  edgelist,
  rem_formula,
  base_riskset = NULL,
  covariates = NULL
)
}
\arguments{
\item{edgelist}{Prepared edgelist (output of \code{\link[=prepare_edgelist]{prepare_edgelist()}}).}

\item{rem_formula}{Survival formula — used to identify covariate columns.}

\item{base_riskset}{Precomputed base risk set; constructed from \code{edgelist}
if \code{NULL}.}

\item{covariates}{Optional exogenous covariate list from \code{\link[=rem_covariates]{rem_covariates()}}.}
}
\value{
Named numeric vector: mean of each REM covariate across events.
}
\description{
Builds covariate values only for the realized events in \code{edgelist}
(n_nonevents = 0L) and returns their column means.  Much faster than
\code{\link[=fit_rem]{fit_rem()}} since no Cox model is fitted and the risk set contains exactly
one row per event.  Use to measure convergence in sufficient-statistic space
rather than coefficient space.
}
\examples{
data(edgelist)
el <- prepare_edgelist(edgelist)
compute_seq_stat_means(
  el[1:100, ],
  survival::Surv(time1, occurred) ~ Repetition + RecRecencyLog +
    survival::strata(time2)
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/providers.R
\name{llm_provider_gemini}
\alias{llm_provider_gemini}
\title{Create a Google Gemini LLM provider}
\usage{
llm_provider_gemini(
  api_key = Sys.getenv("GEMINI_API_KEY"),
  model = "gemini-2.0-flash",
  max_retries = 5L
)
}
\arguments{
\item{api_key}{Defaults to \code{Sys.getenv("GEMINI_API_KEY")}.}

\item{model}{Default \code{"gemini-2.0-flash"}.}

\item{max_retries}{Maximum request attempts.}
}
\value{
An \code{llm_provider} object.
}
\description{
Uses Google's OpenAI-compatible endpoint.
}
\examples{
\dontrun{
p <- llm_provider_gemini()   # reads GEMINI_API_KEY from env
print(p)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rem.R
\name{fit_rem}
\alias{fit_rem}
\title{Fit a Relational Event Model to an edgelist}
\usage{
fit_rem(
  edgelist,
  formula,
  n = NULL,
  n_nonevents = 200L,
  base_riskset = NULL,
  cl = NULL,
  method = c("cox", "clogit", "temporal"),
  covariates = NULL
)
}
\arguments{
\item{edgelist}{Data frame with columns \code{time}, \code{sender}, \code{receiver},
\code{eveid}. Use \code{\link[=prepare_edgelist]{prepare_edgelist()}} to create this from a raw matrix.}

\item{formula}{A \code{survival::Surv} formula. Must reference \code{time1} and
\code{time2} (via \code{strata()}).

\strong{Available covariates}

\tabular{llll}{
\strong{Name} \tab \strong{Definition} \tab \strong{relevent analogue} \tab \strong{Notes} \cr
\emph{Dyad activity} \tab \tab \tab \cr
\code{Repetition} \tab dyad_occ / sender_outDeg \tab \code{FrPSndSnd} \tab Same quantity; llrem denom = n-1 (fixed), relevent = event index \cr
\code{SndInitiation} \tab I(dyad i→j has ever fired) \tab — \tab llrem only; used as partition variable in \code{\link[=fit_rem_sep]{fit_rem_sep()}} \cr
\code{DyadRecencyLog} \tab -log1p(time since last i→j) \tab \code{RSndSnd} \tab Same direction; relevent uses recency rank, not log-time \cr
\code{WeightedRep} \tab w_ij / max(w_ij in risk set) \tab — \tab llrem only; same-dyad count normalised by per-sender max \cr
\emph{Reciprocity} \tab \tab \tab \cr
\code{RRecInitiation} \tab I(reverse dyad j→i has ever fired) \tab \code{PSAB-BA} (one-step) \tab relevent PSAB-BA resets after each non-BA event; RRecInitiation is permanent \cr
\code{RecRecencyLog} \tab -log1p(time since last j→i) \tab \code{RRecSnd} \tab Both measure reverse-dyad recency; relevent uses rank, llrem uses log-time \cr
\code{RecRecencyRank} \tab 1 / rank(j in i's incoming recency list) \tab \code{RRecSnd} \tab Exact relevent analogue; rank 1 = most recent sender to i \cr
\code{WeightedRec} \tab min(w_ij, w_ji) / max(w_ij, w_ji) \tab — \tab Squartini et al. (2013) weighted reciprocity; 0 if both zero \cr
\code{FrRecSnd} \tab w_ji / sender_inDeg \tab \code{FrRecSnd} \tab Fraction of sender's in-events that are from receiver \cr
\emph{Degree} \tab \tab \tab \cr
\code{OutDegSnd} \tab sender out-degree / (n-1) \tab \code{NODSnd} \tab relevent divides by running event count \cr
\code{OutDegRec} \tab receiver out-degree / (n-1) \tab \code{NODRec} \tab same normalisation difference \cr
\code{InDegSnd} \tab sender in-degree / (n-1) \tab \code{NIDSnd} \tab same normalisation difference \cr
\code{InDegRec} \tab receiver in-degree / (n-1) \tab \code{NIDRec} \tab same normalisation difference \cr
\code{TotDegSnd} \tab (OutDegSnd + InDegSnd) / 2 \tab \code{NTDegSnd} \tab both are total-degree averages \cr
\code{TotDegRec} \tab (OutDegRec + InDegRec) / 2 \tab \code{NTDegRec} \tab both are total-degree averages \cr
\emph{Triadic closure} \tab \tab \tab \cr
\code{OTP} \tab out-two-path count \tab \code{OTPSnd} \tab raw count; use \code{NOTPSnd} for normalised \cr
\code{ITP} \tab in-two-path count \tab \code{ITPSnd} \tab raw count; use \code{NITPSnd} for normalised \cr
\code{OSP} \tab out-shared-partner count \tab \code{OSPSnd} \tab raw count; use \code{NOSPSnd} for normalised \cr
\code{ISP} \tab in-shared-partner count \tab \code{ISPSnd} \tab raw count; use \code{NISPSnd} for normalised \cr
\code{NOTPSnd} \tab OTP / max(OTP in risk set) \tab — \tab llrem only (normalised OTP) \cr
\code{NITPSnd} \tab ITP / max(ITP in risk set) \tab — \tab llrem only (normalised ITP) \cr
\code{NOSPSnd} \tab OSP / max(OSP in risk set) \tab — \tab llrem only (normalised OSP) \cr
\code{NISPSnd} \tab ISP / max(ISP in risk set) \tab — \tab llrem only (normalised ISP) \cr
\code{NTriadic} \tab (OTP + ITP + OSP + ISP) / max(OTP + ITP + OSP + ISP) \tab — \tab llrem only; omnibus triadic score; ranges 0--1 \cr
\emph{Imitation} \tab \tab \tab \cr
\code{Imitation} \tab (recInDg - w_ij) / max per sender \tab — \tab Times j targeted by actors other than i, normalised; ranges 0--1 \cr
\code{PreImitation} \tab recInDg - w_ij (raw, unnormalised) \tab — \tab Raw count underlying Imitation \cr
\emph{P-shifts} \tab \tab \tab \cr
\code{IOviolation} \tab PSAB_XA + PSAB_XB + PSAB_XY (any turn-usurp) \tab \code{PSAB-XA}/\code{PSAB-XB}/\code{PSAB-XY} (separate) \tab llrem collapses the three usurping types \cr
\code{PSAB_BA} \tab I(last event was j→i, this is i→j) \tab \code{PSAB-BA} \tab exact match \cr
\code{PSAB_BY} \tab I(last event was j→i, this is i→y, y!=j) \tab \code{PSAB-BY} \tab exact match \cr
\code{PSAB_AY} \tab I(last event was i→j, this is i→y, y!=j) \tab \code{PSAB-AY} \tab exact match \cr
\code{PSAB_XA} \tab I(last event was x→i, this is j→i) \tab \code{PSAB-XA} \tab exact match \cr
\code{PSAB_XB} \tab I(last event was x→j, this is i→j) \tab \code{PSAB-XB} \tab exact match \cr
\code{PSAB_XY} \tab I(last event was x→y, this is i→j; x!=i, y!=j) \tab \code{PSAB-XY} \tab exact match \cr
}}

\item{n}{Number of nodes. Inferred from \code{edgelist} if \code{NULL}.}

\item{n_nonevents}{Sampled non-event risk-set size per event. \code{NULL} uses
the full risk set (slow for large networks; 200 is a sensible default).}

\item{base_riskset}{Precomputed base risk-set data frame. Constructed from
\code{n} if \code{NULL}. Pass a cached object when calling \code{fit_rem()} and
\code{\link[=run_llm_rem]{run_llm_rem()}} in the same session to avoid redundant work.}

\item{cl}{Optional PSOCK cluster from \code{parallel::makeCluster()} for
parallelised risk-set computation.}

\item{method}{Estimation method: \code{"cox"} (default) fits a Cox proportional
hazards model via \code{\link[survival:coxph]{survival::coxph()}} with one stratum per event.
\code{"clogit"} fits a conditional logit (discrete-choice/ordinal) model via
BFGS; substantially faster than \code{"cox"} for large sequences and
asymptotically equivalent under the ordinal likelihood.
\code{"temporal"} fits the full interval-censored REM likelihood, which adds a
waiting-time penalty \verb{Delta_t * Sigma exp(eta)} to each event's contribution.
Requires that \code{edgelist$time} contains real inter-event gaps, not just
ranks; coefficients are no longer scale-free.

Available covariates: see \verb{@param formula} above.

Note on naming vs \strong{relevent}: \code{OutDegSnd}/\code{OutDegRec}/\code{InDegSnd}/\code{InDegRec}
scale by \code{(n-1)} (fixed maximum degree), unlike relevent's \code{NODSnd}/\code{NIDRec}
which scale by the running event count.  \code{DyadRecencyLog} is
\verb{-log1p(clock-time of last i→j event)}; relevent's \code{RSndSnd} is a recency
rank -- both measure same-dyad recency but are not numerically comparable.
\code{RecRecencyLog} (\verb{-log1p(time since last j→i)}) and \code{RecRecencyRank}
(\code{1/rank} of j in i's incoming recency list) are both analogues of
relevent's \code{RRecSnd}; \code{RecRecencyRank} is the exact functional equivalent.}

\item{covariates}{Optional exogenous covariate list created by
\code{\link[=rem_covariates]{rem_covariates()}}. \code{NULL} uses only the built-in endogenous statistics.}
}
\value{
For \code{method = "cox"}: a \code{coxph} object.  For \code{method = "clogit"} or
\code{"temporal"}: a \code{rem_clogit} object with elements \code{coef}, \code{vcov},
\code{loglik}, and \code{converged}.
}
\description{
Builds risk-set statistics for each event in the sequence, adds survival
model timing columns (\code{time1}, \code{time2}), and fits a Cox proportional hazards
model. All covariate computation is handled internally.
}
\examples{
data(edgelist)
el <- prepare_edgelist(edgelist)
fit <- fit_rem(
  el[1:100, ],
  survival::Surv(time1, occurred) ~ Repetition + RecRecencyLog +
    survival::strata(time2),
  n_nonevents = 20L, method = "clogit"
)
print(fit)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/providers.R
\name{llm_provider_anthropic}
\alias{llm_provider_anthropic}
\title{Create an Anthropic Claude LLM provider}
\usage{
llm_provider_anthropic(
  api_key = Sys.getenv("ANTHROPIC_API_KEY"),
  model = "claude-sonnet-4-6",
  anthropic_version = "2023-06-01",
  max_retries = 5L
)
}
\arguments{
\item{api_key}{Defaults to \code{Sys.getenv("ANTHROPIC_API_KEY")}.}

\item{model}{Default \code{"claude-3-5-sonnet-20241022"}.}

\item{anthropic_version}{API version header string.}

\item{max_retries}{Maximum request attempts.}
}
\value{
An \code{llm_provider} object.
}
\description{
Anthropic's API separates the system prompt as a top-level field and uses a
different response shape than the OpenAI format. This constructor handles
both translations internally.
}
\examples{
\dontrun{
p <- llm_provider_anthropic()   # reads ANTHROPIC_API_KEY from env
print(p)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rem.R
\name{run_llm_rem}
\alias{run_llm_rem}
\title{Generate an LLM event sequence and fit a REM}
\usage{
run_llm_rem(
  edgelist,
  formula,
  query_fn,
  n_events = NULL,
  n = NULL,
  n_nonevents = 200L,
  base_riskset = NULL,
  cl = NULL,
  seed = 8214L,
  checkpoint_path = NULL,
  use_empirical_history = FALSE,
  rate_limit_secs = 0.3,
  fit = TRUE,
  method = c("cox", "clogit", "temporal"),
  covariates = NULL
)
}
\arguments{
\item{edgelist}{Prepared edgelist (output of \code{\link[=prepare_edgelist]{prepare_edgelist()}}).}

\item{formula}{Survival formula passed to \code{\link[=fit_rem]{fit_rem()}}.}

\item{query_fn}{A closure produced by \code{\link[=make_query_roleplay_sender]{make_query_roleplay_sender()}},
\code{\link[=make_query_predict]{make_query_predict()}}, or \code{\link[=make_query_roleplay_random]{make_query_roleplay_random()}}, or any function
with the signature:
\code{f(past_events, empirical_event, all_nodes) -> c(sender, receiver)}.}

\item{n_events}{Number of events to generate. \code{NULL} uses the full edgelist.}

\item{n}{Number of nodes. Inferred if \code{NULL}.}

\item{n_nonevents}{Sampled risk-set size passed to \code{\link[=fit_rem]{fit_rem()}}.}

\item{base_riskset}{Precomputed base risk set.}

\item{cl}{Optional PSOCK cluster for parallelised risk-set computation.}

\item{seed}{RNG seed for random fallbacks and random-node sampling.}

\item{checkpoint_path}{File path for incremental saves every 10 events.
\code{NULL} disables checkpointing.}

\item{use_empirical_history}{If \code{FALSE} (default), the LLM is shown its own
previously generated events as history, so any divergence from reality
compounds over time. If \code{TRUE}, the LLM is shown the empirical event
history at each step, isolating the effect of the generation strategy from
the effect of accumulated sequence drift.}

\item{rate_limit_secs}{Seconds to pause between LLM calls.}

\item{fit}{If \code{TRUE} (default), fits a REM on the generated sequence before
returning. Set to \code{FALSE} to skip model fitting and return only the
sequence and raw responses — useful when running many sequences and
deferring all model fitting to a separate step.}

\item{method}{Passed to \code{\link[=fit_rem]{fit_rem()}}: \code{"cox"} (default), \code{"clogit"}, or
\code{"temporal"}.}

\item{covariates}{Optional exogenous covariate list passed to \code{\link[=fit_rem]{fit_rem()}}.}
}
\value{
A named list with elements:
\itemize{
\item \code{sequence}: generated event sequence as a data frame.
\item \code{responses}: raw LLM response text and parsed sender/receiver per event.
\item \code{rem}: fitted model (only present when \code{fit = TRUE}).
}
}
\description{
At each time step \emph{t} (starting from 2), \code{query_fn} is called with the
growing LLM-generated history, the empirical event at \emph{t}, and the full
node set. The returned sender/receiver pair is appended to the sequence.
The sequence is seeded with the first empirical event so the LLM starts
with at least one data point. After generation, risk-set statistics are
built and a REM is fitted.
}
\examples{
\dontrun{
data(edgelist)
el  <- prepare_edgelist(edgelist)
cfg <- rem_cfg(
  provider    = llm_provider_openai(),
  context     = "These are militarized conflict events between countries.",
  action_verb = "attacked",
  actor_noun  = "Country"
)
res <- run_llm_rem(
  el[1:50, ],
  survival::Surv(time1, occurred) ~ Repetition + RecRecencyLog +
    survival::strata(time2),
  query_fn    = make_query_roleplay_sender(cfg),
  n_nonevents = 50L, method = "clogit"
)
print(res$rem)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/covariates.R
\name{rem_covariates}
\alias{rem_covariates}
\title{Collect covariate objects for use in fit_rem}
\usage{
rem_covariates(...)
}
\arguments{
\item{...}{Any number of objects created by \code{\link[=cov_node_static]{cov_node_static()}},
\code{\link[=cov_node_temporal]{cov_node_temporal()}}, \code{\link[=cov_dyad_static]{cov_dyad_static()}}, or \code{\link[=cov_dyad_temporal]{cov_dyad_temporal()}}.}
}
\value{
An object of class \code{"rem_covariates"}.
}
\description{
Validates that all covariate names are unique, do not collide with
endogenous REM column names, and are proper \code{rem_cov} objects.
}
\examples{
gdp  <- cov_node_static("GDP", data.frame(node = 1:4, value = c(1,2,1,3)))
ally <- cov_dyad_static("Alliance",
          data.frame(sender = 1L, receiver = 2L, value = 1))
covs <- rem_covariates(gdp, ally)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/providers.R
\name{llm_provider_openai_compat}
\alias{llm_provider_openai_compat}
\title{Create a generic OpenAI-compatible LLM provider}
\usage{
llm_provider_openai_compat(
  api_key,
  model,
  base_url,
  name = "OpenAI-compatible",
  max_retries = 5L
)
}
\arguments{
\item{api_key}{API key string.}

\item{model}{Model identifier.}

\item{base_url}{Full chat-completions endpoint URL.}

\item{name}{Display name used in \code{print()} output and plot labels.}

\item{max_retries}{Maximum request attempts before returning \code{NA}.}
}
\value{
An \code{llm_provider} object.
}
\description{
Suitable for any API that mirrors the OpenAI chat completions wire format.
Used internally by \code{\link[=llm_provider_openai]{llm_provider_openai()}}, \code{\link[=llm_provider_grok]{llm_provider_grok()}},
\code{\link[=llm_provider_gemini]{llm_provider_gemini()}}, and \code{\link[=llm_provider_ollama]{llm_provider_ollama()}}.
}
\examples{
\dontrun{
p <- llm_provider_openai_compat(
  api_key  = Sys.getenv("OPENAI_API_KEY"),
  model    = "gpt-4o-mini",
  base_url = "https://api.openai.com/v1/chat/completions"
)
print(p)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multiagent.R
\name{compute_hazard_surface}
\alias{compute_hazard_surface}
\title{Compute the full hazard surface over all directed dyads}
\usage{
compute_hazard_surface(history, base_riskset, beta, covariates = NULL)
}
\arguments{
\item{history}{Data frame of past events (columns: time, sender,
receiver, eveid).}

\item{base_riskset}{Precomputed base risk set (output of \code{get_base_riskset()}).}

\item{beta}{Named numeric vector of REM coefficients (from
\code{coef(rem_fit)}).}

\item{covariates}{Optional exogenous covariate list from \code{\link[=rem_covariates]{rem_covariates()}}.}
}
\value{
Data frame with columns \code{sender}, \code{receiver}, \code{hazard}.
}
\description{
Evaluates \code{exp(beta^T * x_ij)} for every directed dyad given the current
event history. Used internally by \code{\link[=run_multiagent_rem]{run_multiagent_rem()}} to derive sender
activation probabilities and the REM-hazard fallback for failed LLM parses.
}
\examples{
data(edgelist)
el   <- prepare_edgelist(edgelist)
fit  <- fit_rem(el[1:100, ],
  survival::Surv(time1, occurred) ~ Repetition + survival::strata(time2),
  n_nonevents = 20L, method = "clogit")
nodes  <- sort(unique(c(el$sender, el$receiver)))
rs     <- expand.grid(sender = nodes, receiver = nodes)
rs     <- rs[rs$sender != rs$receiver, ]
rs$eid <- paste0(rs$sender, "_", rs$receiver)
beta <- c(Repetition = coef(fit)[["Repetition"]])
hz   <- compute_hazard_surface(el[1:5, ], rs, beta)
head(hz)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/providers.R
\name{llm_call}
\alias{llm_call}
\title{Call an LLM provider}
\usage{
llm_call(provider, messages, max_tokens = 15L, temperature = 0.7)
}
\arguments{
\item{provider}{An \code{llm_provider} object.}

\item{messages}{List of role/content pairs in OpenAI format, e.g.
\code{list(list(role = "system", content = "..."), list(role = "user", content = "..."))}.}

\item{max_tokens}{Token budget for the response.}

\item{temperature}{Sampling temperature.}
}
\value{
A character string, or \code{NA_character_} on failure.
}
\description{
Generic dispatch function. Passes a list of role/content message pairs to
the provider's internal call function and returns the text response.
}
\examples{
\dontrun{
p    <- llm_provider_openai()
msgs <- list(
  list(role = "system", content = "Reply with a single integer."),
  list(role = "user",   content = "What is 2 + 2?")
)
llm_call(p, msgs, max_tokens = 5L)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/covariates.R
\name{cov_dyad_static}
\alias{cov_dyad_static}
\title{Create a static dyad-level covariate}
\usage{
cov_dyad_static(name, data, warn_missing = TRUE)
}
\arguments{
\item{name}{Character scalar. Column name added to the risk set.}

\item{data}{Data frame with columns \code{sender} (integer-coercible),
\code{receiver} (integer-coercible), and \code{value} (numeric). Need not be a
complete cross-product; unmatched dyads receive 0 (with optional warning).}

\item{warn_missing}{Logical.}
}
\value{
An object of class \code{c("cov_dyad_static", "rem_cov")}.
}
\description{
Adds one column to each risk-set row based on the (sender, receiver) pair.
}
\examples{
alliance <- data.frame(
  sender = c(1L, 2L), receiver = c(2L, 1L), value = c(1, 1)
)
cov <- cov_dyad_static("Alliance", alliance)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/covariates.R
\name{cov_node_static}
\alias{cov_node_static}
\title{Create a static node-level covariate}
\usage{
cov_node_static(
  name,
  data,
  role = c("both", "sender", "receiver"),
  warn_missing = TRUE
)
}
\arguments{
\item{name}{Character scalar. Column name(s) added to the risk set.
When \code{role = "both"}, two columns are added: \verb{<name>_snd} and
\verb{<name>_rec}. Otherwise a single column named \verb{<name>} is added.}

\item{data}{Data frame with columns \code{node} (integer-coercible) and
\code{value} (numeric). One row per node.}

\item{role}{One of \code{"both"} (default), \code{"sender"}, or \code{"receiver"}.}

\item{warn_missing}{Logical. Warn when a node in the risk set has no
matching row in \code{data} (default \code{TRUE}).}
}
\value{
An object of class \code{c("cov_node_static", "rem_cov")}.
}
\description{
Adds one or two columns to each risk-set row based on the sender's and/or
receiver's value in a node attribute table.
}
\examples{
gdp <- data.frame(node = 1:4, value = c(1.2, 0.8, 2.1, 0.5))
cov <- cov_node_static("GDP", gdp, role = "sender")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock.R
\name{mock_strategy_max_id}
\alias{mock_strategy_max_id}
\title{Mock strategy: always pick the largest valid target}
\usage{
mock_strategy_max_id(messages, max_tokens, temperature)
}
\arguments{
\item{messages}{List of message objects passed by the provider interface.}

\item{max_tokens}{Maximum tokens (ignored by mock; present for interface
compatibility).}

\item{temperature}{Sampling temperature (ignored by mock; present for
interface compatibility).}
}
\value{
A strategy function for use with \code{\link[=llm_provider_mock]{llm_provider_mock()}}.
}
\description{
Mock strategy: always pick the largest valid target
}
\examples{
msgs <- list(list(role = "user",
  content = "Valid targets: (2) (3) (4)"))
mock_strategy_max_id(msgs, 15L, 0.7)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prompts.R
\name{make_query_roleplay_sender}
\alias{make_query_roleplay_sender}
\title{Query factory: LLM roleplays as the empirical sender}
\usage{
make_query_roleplay_sender(cfg)
}
\arguments{
\item{cfg}{A \code{rem_cfg} object.}
}
\value{
A function with signature
\code{f(past_events, empirical_event, all_nodes) -> c(sender, receiver)}.
}
\description{
Returns a \code{query_fn} closure in which the LLM is told it IS the empirical
sender at time \emph{t} and asked to choose a receiver. The generated sequence
preserves the empirical sender ordering but uses LLM-selected receivers.
}
\examples{
\dontrun{
cfg <- rem_cfg(
  provider    = llm_provider_openai(),
  context     = "Conflict events between states.",
  action_verb = "attacked",
  actor_noun  = "Country"
)
query_fn <- make_query_roleplay_sender(cfg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_correction_query.R
\name{make_suffstats}
\alias{make_suffstats}
\title{Extract sufficient-statistic means from a REM risk-set data frame}
\usage{
make_suffstats(rem_formula, data, drop_strata = TRUE)
}
\arguments{
\item{rem_formula}{Survival formula whose RHS terms are the covariates.}

\item{data}{Risk-set data frame (expanded counting-process format)
with an \code{occurred} column.}

\item{drop_strata}{If \code{TRUE} (default), drop strata columns from the model
matrix before computing means.}
}
\value{
A data frame with columns \code{occurred} plus one column per covariate.
}
\description{
Returns a data frame with one row per value of \code{occurred} (0/1), containing
the mean of each REM covariate.  Used as input to \code{\link[=make_behavior_queries]{make_behavior_queries()}}.
}
\examples{
data(edgelist)
el  <- prepare_edgelist(edgelist)
fit <- fit_rem(el[1:100, ],
  survival::Surv(time1, occurred) ~ Repetition + survival::strata(time2),
  n_nonevents = 30L, method = "clogit")
# make_suffstats operates on the expanded risk-set data frame; see fit_rem
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multiagent.R
\name{run_multiagent_rem}
\alias{run_multiagent_rem}
\title{Run a multi-agent REM-LLM conflict simulation}
\usage{
run_multiagent_rem(
  edgelist,
  rem_fit,
  cfg,
  n_ticks,
  lambda = NULL,
  base_riskset = NULL,
  cold_start = FALSE,
  seed = 8214L,
  checkpoint_path = NULL,
  covariates = NULL
)
}
\arguments{
\item{edgelist}{Prepared edgelist -- used to infer the node set and
optionally as the seed event.}

\item{rem_fit}{Fitted \code{coxph} object (from \code{\link[=fit_rem]{fit_rem()}}).}

\item{cfg}{A single \code{\link[=rem_cfg]{rem_cfg()}} shared across all agents.
Sender identity is injected dynamically at each query.}

\item{n_ticks}{Number of simulation ticks.}

\item{lambda}{Mean events per tick (Poisson rate). Estimated from
\code{edgelist} inter-event times if \code{NULL}.}

\item{base_riskset}{Precomputed base risk set. Built from \code{edgelist} if
\code{NULL}.}

\item{cold_start}{If \code{TRUE}, begin with a uniform hazard surface rather
than seeding from the first empirical event. Cold-start runs have a burn-in
period before endogenous structure accumulates.}

\item{seed}{RNG seed.}

\item{checkpoint_path}{File path for incremental saves every 10 ticks.
\code{NULL} disables checkpointing. Existing checkpoint is resumed automatically.}

\item{covariates}{Optional exogenous covariate list from \code{\link[=rem_covariates]{rem_covariates()}}.
Passed to \code{\link[=fit_rem]{fit_rem()}} when computing the hazard surface.}
}
\value{
A named list:
\describe{
\item{\code{sequence}}{Data frame of generated events (time, sender, receiver,
eveid).}
\item{\code{sender_probs}}{\code{n_ticks} x \code{N} matrix of activation probabilities
\eqn{p_i(t)} at each tick.}
\item{\code{n_events_per_tick}}{Integer vector of \eqn{K_t} values.}
}
}
\description{
At each tick, the REM hazard surface determines who acts (sender activation),
and individual LLM agents decide against whom (receiver selection). The two
components are complementary: the REM enforces endogenous structure; the LLM
contributes identity-conditioned content.
}
\examples{
\dontrun{
data(edgelist)
el  <- prepare_edgelist(edgelist)
fit <- fit_rem(el[1:200, ],
  survival::Surv(time1, occurred) ~ Repetition + Recencylog +
    survival::strata(time2),
  n_nonevents = 50L, method = "cox")
cfg <- rem_cfg(
  provider    = llm_provider_openai(),
  context     = "Conflict events between states.",
  action_verb = "attacked",
  actor_noun  = "Country"
)
sim <- run_multiagent_rem(el[1:200, ], fit, cfg, n_ticks = 5L)
head(sim$sequence)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/providers.R
\name{llm_provider_bedrock}
\alias{llm_provider_bedrock}
\title{Create an AWS Bedrock LLM provider}
\usage{
llm_provider_bedrock(
  model,
  region = Sys.getenv("AWS_DEFAULT_REGION", unset = "us-east-1"),
  access_key = Sys.getenv("AWS_ACCESS_KEY_ID"),
  secret_key = Sys.getenv("AWS_SECRET_ACCESS_KEY"),
  session_token = Sys.getenv("AWS_SESSION_TOKEN"),
  name = "AWS Bedrock",
  max_retries = 5L
)
}
\arguments{
\item{model}{Bedrock model ID string (see above).}

\item{region}{AWS region. Defaults to \code{AWS_DEFAULT_REGION} env var,
then \code{"us-east-1"}.}

\item{access_key}{AWS access key ID. Defaults to \code{AWS_ACCESS_KEY_ID}.}

\item{secret_key}{AWS secret access key. Defaults to \code{AWS_SECRET_ACCESS_KEY}.}

\item{session_token}{Optional STS session token. Defaults to
\code{AWS_SESSION_TOKEN} (ignored if unset).}

\item{name}{Display name for print output and plot labels.}

\item{max_retries}{Maximum request attempts.}
}
\value{
An \code{llm_provider} object.
}
\description{
Uses Bedrock's Converse API with AWS SigV4 signing. Supports any model
available in your Bedrock account, including Anthropic Claude and Meta
Llama families.
}
\details{
Credentials are resolved in order: explicit arguments → environment variables
(\code{AWS_ACCESS_KEY_ID}, \code{AWS_SECRET_ACCESS_KEY}, \code{AWS_SESSION_TOKEN},
\code{AWS_DEFAULT_REGION}).

Common model IDs:
\itemize{
\item \code{"anthropic.claude-haiku-4-5-20251001-v1:0"}
\item \code{"anthropic.claude-3-5-sonnet-20241022-v2:0"}
\item \code{"meta.llama3-70b-instruct-v1:0"}
\item \code{"meta.llama3-8b-instruct-v1:0"}
\item \code{"mistral.mistral-7b-instruct-v0:2"}
}
}
\examples{
\dontrun{
p <- llm_provider_bedrock(
  model = "anthropic.claude-haiku-4-5-20251001-v1:0"
)   # reads AWS_* credentials from env
print(p)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock.R
\name{llm_provider_mock}
\alias{llm_provider_mock}
\title{Create a mock LLM provider for testing}
\usage{
llm_provider_mock(strategy_fn, name = "Mock")
}
\arguments{
\item{strategy_fn}{Function with signature
\code{f(messages, max_tokens, temperature) -> character}. The return value
must be parseable by the corresponding query factory (a stringified
integer for roleplay scenarios; two space-separated integers for the
predict scenario).}

\item{name}{Display name for \code{print()} output.}
}
\value{
An \code{llm_provider} object.
}
\description{
Returns an \code{llm_provider} that calls a local \code{strategy_fn} instead of
making any network requests. Useful for testing pipeline logic and custom
\code{query_fn} closures without incurring API costs.
}
\details{
Three ready-made strategies are provided:
\itemize{
\item \code{\link[=mock_strategy_min_id]{mock_strategy_min_id()}}: always picks the smallest valid target ID.
\item \code{\link[=mock_strategy_max_id]{mock_strategy_max_id()}}: always picks the largest valid target ID.
\item \code{\link[=mock_strategy_highest_indegree]{mock_strategy_highest_indegree()}}: picks the valid target with the
highest cumulative in-degree in the event history.
}
}
\examples{
p <- llm_provider_mock(mock_strategy_min_id)
print(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/riskset.R
\name{prepare_edgelist}
\alias{prepare_edgelist}
\title{Normalise a raw edgelist to the standard three-column format}
\usage{
prepare_edgelist(raw)
}
\arguments{
\item{raw}{Matrix or data frame with at least 3 columns.}
}
\value{
A data frame with columns \code{time} (integer), \code{sender} (integer),
\code{receiver} (integer), and \code{eveid} (character).
}
\description{
Accepts a matrix or data frame with columns in order: time, sender,
receiver. Returns a data frame with named integer columns and an \code{eveid}
dyad key (\code{"sender_receiver"}).
}
\examples{
el <- prepare_edgelist(data.frame(t = 1:5, s = c(1,2,3,1,2), r = c(2,3,1,3,1)))
head(el)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prompts.R
\name{make_query_predict}
\alias{make_query_predict}
\title{Query factory: LLM predicts the next event}
\usage{
make_query_predict(cfg)
}
\arguments{
\item{cfg}{A \code{rem_cfg} object.}
}
\value{
A function with signature
\code{f(past_events, empirical_event, all_nodes) -> c(sender, receiver)}.
}
\description{
Returns a \code{query_fn} closure that shows the LLM the event history and asks
it to predict both the sender and receiver of the next event. Tests the
model's implicit understanding of network dynamics rather than its own
behavioural tendencies.
}
\examples{
\dontrun{
cfg <- rem_cfg(
  provider    = llm_provider_openai(),
  context     = "Conflict events between states.",
  action_verb = "attacked",
  actor_noun  = "Country"
)
query_fn <- make_query_predict(cfg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/providers.R
\name{llm_provider_ollama}
\alias{llm_provider_ollama}
\title{Create a local or remote Ollama LLM provider}
\usage{
llm_provider_ollama(
  model = "llama3.2",
  host = "localhost",
  port = 11434L,
  base_url = NULL,
  max_retries = 3L
)
}
\arguments{
\item{model}{Model tag, e.g. \code{"llama3.2"} or \code{"mistral"}.}

\item{host}{Hostname or IP of the Ollama server.  Defaults to
\code{"localhost"}.  Override to target a remote machine, e.g.
\code{host = "192.168.1.65"}.}

\item{port}{Port the Ollama server is listening on (default 11434).}

\item{base_url}{Full endpoint URL.  When supplied, \code{host} and \code{port} are
ignored — useful as a manual escape hatch.}

\item{max_retries}{Maximum request attempts.}
}
\value{
An \code{llm_provider} object.
}
\description{
Create a local or remote Ollama LLM provider
}
\examples{
\dontrun{
p <- llm_provider_ollama(model = "llama3.2")   # requires local Ollama server
print(p)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{compare_rem_models}
\alias{compare_rem_models}
\title{Forest plot comparing fitted REM coefficients}
\usage{
compare_rem_models(
  models,
  term_order = NULL,
  palette = NULL,
  title = "REM Coefficient Comparison"
)
}
\arguments{
\item{models}{Named list of \code{coxph} objects.}

\item{term_order}{Character vector giving the desired y-axis order (bottom
to top). Defaults to reverse alphabetical over all terms present.}

\item{palette}{Named character vector of hex colours keyed to
\code{names(models)}. A colour-blind-friendly default palette is applied when
\code{NULL}.}

\item{title}{Plot title string.}
}
\value{
A \code{ggplot} object (invisibly). The plot is also printed as a
side-effect.
}
\description{
Produces a coefficient plot with 95\% confidence intervals for a named list
of \code{coxph} models, with one colour per model source.
}
\examples{
\dontrun{
data(edgelist)
el <- prepare_edgelist(edgelist)
f  <- survival::Surv(time1, occurred) ~ Repetition + Recencylog +
         survival::strata(time2)
m1 <- fit_rem(el[1:150, ], f, n_nonevents = 30L, method = "cox")
m2 <- fit_rem(el[1:300, ], f, n_nonevents = 30L, method = "cox")
compare_rem_models(list("N=150" = m1, "N=300" = m2))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/covariates.R
\name{attach_covariates_df}
\alias{attach_covariates_df}
\title{Attach exogenous covariates to a pre-built risk-set data frame}
\usage{
attach_covariates_df(pe, covariates)
}
\arguments{
\item{pe}{Data frame returned by \code{remstats_generate()} (or
equivalently, the \code{pe} object inside \code{\link[=fit_rem]{fit_rem()}}), with a \code{time} column.}

\item{covariates}{A \code{\link[=rem_covariates]{rem_covariates()}} object.}
}
\value{
\code{pe} with additional covariate columns appended.
}
\description{
Convenience wrapper for applying a \code{\link[=rem_covariates]{rem_covariates()}} object to the output
of \code{remstats_generate()} after the fact.  Useful for testing or for
workflows that build the risk set once and experiment with different
covariate specifications.
}
\examples{
\dontrun{
data(edgelist)
el   <- prepare_edgelist(edgelist)
gdp  <- cov_node_static("GDP", data.frame(
  node = sort(unique(c(el$sender, el$receiver))),
  value = runif(length(unique(c(el$sender, el$receiver))))
))
covs <- rem_covariates(gdp)
fit  <- fit_rem(el[1:100, ],
  survival::Surv(time1, occurred) ~ GDP + Repetition +
    survival::strata(time2),
  n_nonevents = 30L, method = "clogit", covariates = covs)
}
}
